#!python
# Copyright (c) 2003 ActiveState Corporation.
# See the file LICENSE.txt for licensing information.

"""The backend for Komodo's new new Rx windows (aka Rxx)."""
#TODO:
# - PERF: Asynchronous matching and reporting of results so that regexes with
#   thousands of matches are not too slow.
# - PERF: delayed searching (similar to linting), so that updating is not
#   attempted more than once per second.

import os
import sys
import string
import re
import logging
import types
import pprint

from xpcom import components, nsError, ServerException, COMException
from xpcom._xpcom import PROXY_SYNC, PROXY_ALWAYS, PROXY_ASYNC, getProxyForObject
from xpcom.server import UnwrapObject
from koTreeView import TreeView

import lexregex


#---- globals

log = logging.getLogger("koRxx")
#log.setLevel(logging.DEBUG)


#---- Rxx backend

class KoRxxMatchResultsTreeView(TreeView):
    _com_interfaces_ = [components.interfaces.koIRxxMatchResultsTreeView]
    _reg_clsid_ = "{882803F0-A674-43EB-9F2C-D82009394719}"
    _reg_contractid_ = "@activestate.com/koRxxMatchResultsTreeView;1"
    _reg_desc_ = "Komodo Rxx Match Results Tree View"

    def __init__(self):
        TreeView.__init__(self) #, debug="rxx")
        self._tree = None
        self._sortedBy = None
        self._matches = None
        # Maps match number to a boolean indicating if the match's twisty
        # is opened. No entry for a match number means the match is _open_.
        # This is maintained across .setMatches() calls so there is no
        # jitter in the match results pane.
        self._matchIsOpen = {}
        self._rows = None

    def setMatches(self, matches):
        """Set the matches for this pattern/searchText.
        
            "matches" is a list of matches. Each match is a list of dicts
                of the form:
                    {"name": <group name or None>,
                     "span": <span>, # (<start index>, <end index>)
                     "value": <value>,
                     "replacement": <replacement>}  # iff a replace result
                There should always be an implicit first group (0) that is
                the whole match.
        """
        #pprint.pprint(matches)
        self._matches = matches
        self._tree.beginUpdateBatch()
        self._updateRows()
        self._tree.invalidate()
        self._tree.endUpdateBatch()

    def _getCurrentMatchDatum(self, datumName):
        index = self._tree.selection.currentIndex
        if index == -1:
            raise ServerException(nsError.NS_ERROR_FAILURE,
                                  "no currently selected match result")
        return self._rows[index][datumName]

    def get_currentMatchGroupName(self):
        return self._getCurrentMatchDatum("group-name")

    def get_currentMatchStart(self):
        return self._getCurrentMatchDatum("group-span")[0]

    def get_currentMatchEnd(self):
        return self._getCurrentMatchDatum("group-span")[1]

    def get_currentMatchValue(self):
        return self._getCurrentMatchDatum("match-results-tree-value")

    def _updateRows(self):
        """Update self._rows based on the current set of matches and on
        which matches are opened.
        """
        rows = []
        for i in range(len(self._matches)):
            groups = self._matches[i]
            row = {"match-results-tree-group": "Match %d (Group 0)" % i,
                   "match-results-tree-span": "%s-%s" % groups[0]["span"],
                   "match-results-tree-value": groups[0]["value"],
                   "match-results-tree-replacement": groups[0].get("replacement", ""),
                   "group-name": None,
                   "group-span": groups[0]["span"],
                   "is-container": len(groups) > 1,
                   "level": 0,
                   "parent-index": -1,
                   "match-index": i}
            rows.append(row)
            if self._matchIsOpen.get(i, 1):
                parentIndex = len(rows) - 1
                for j in range(1, len(groups)):
                    group = groups[j]
                    if group["value"] is None:
                        continue
                    if group["name"]:
                        name = "Group %d (%s)" % (j, group["name"])
                    else:
                        name = "Group %d" % j
                    row = {"match-results-tree-group": name,
                           "match-results-tree-span": "%s-%s" % group["span"],
                           "match-results-tree-value": group["value"],
                           "match-results-tree-replacement": group.get("replacement", ""),
                           "group-name": group["name"],
                           "group-span": group["span"],
                           "is-container": 0,
                           "level": 1,
                           "parent-index": parentIndex,
                           "match-index": i}
                    rows.append(row)
        #pprint.pprint(rows)
        self._rows = rows

    #---- nsITreeView methods
    def get_rowCount(self):
        try:
            return len(self._rows)
        except TypeError: # self._rows is None
            return 0

    def getCellText(self, row, col):
        cell = ""
        try:
            cell = self._rows[row][col]
        except IndexError:
            log.error("no %sth match result" % row)
        except KeyError:
            log.error("unknown match results column: '%s'" % col)
        if type(cell) not in types.StringTypes:
            cell = str(cell)
        return cell

    def isContainer(self, row):
        return self._rows[row]["is-container"]

    def isContainerEmpty(self, row):
        #XXX Maybe we should use this and always make a match a container
        #    even if there are no sub groups???
        return 0

    def isContainerOpen(self, row):
        matchIndex = self._rows[row]["match-index"]
        return self._matchIsOpen.get(matchIndex, 1)

    def toggleOpenState(self, row):
        matchIndex = self._rows[row]["match-index"]
        isOpen = self._matchIsOpen.get(matchIndex, 1)
        self._matchIsOpen[matchIndex] = not isOpen

        # Must recalculate the rows.
        self._tree.beginUpdateBatch()
        self._updateRows()
        self._tree.invalidate()
        self._tree.endUpdateBatch()

    def getParentIndex(self, row):
        return self._rows[row]["parent-index"]

    def getLevel(self, row):
        return self._rows[row]["level"]

    #XXX might handle this from JS side
    #def selectionChanged(self):
    #    index = self.selection.currentIndex
    #    cdict = self._data[index]
    #    self.templatesView.setData(cdict["node"].files)

    def getImageSrc(self, row, col):
        #XXX Get more appropriate icons
        if col != "match-results-tree-group":
            return ""
        if self._rows[row]["is-container"]:
            matchIndex = self._rows[row]["match-index"]
            isOpen = self._matchIsOpen.get(matchIndex, 1)
            if isOpen:
                return "chrome://jaguar/skin/images/folder-open.png"
            else:
                return "chrome://jaguar/skin/images/folder-closed.png"
        #else:
        #   XXX might want to have an icon for the simple groups
    
    def hasNextSibling(self, rowIndex, afterIndex):
        """From the nsITreeView.idl docs:
        
        HasNextSibling is used to determine if the row at rowIndex has a
        nextSibling that occurs *after* the index specified by
        afterIndex.  Code that is forced to march down the view looking
        at levels can optimize the march by starting at afterIndex+1.
        """
        if afterIndex+1 >= len(self._rows):
            return 0
        else:
            return (self._rows[rowIndex]["parent-index"]
                    == self._rows[afterIndex+1]["parent-index"])


class KoRxxSplitResultsTreeView(TreeView):
    _com_interfaces_ = [components.interfaces.koIRxxSplitResultsTreeView]
    _reg_clsid_ = "{2C199F24-E11E-467C-BA96-9178605FD810}"
    _reg_contractid_ = "@activestate.com/koRxxSplitResultsTreeView;1"
    _reg_desc_ = "Komodo Rxx Split Results Tree View"

    def __init__(self):
        TreeView.__init__(self) #, debug="rxx")
        self._tree = None
        self._rows = []

    def setSplitResults(self, results, searchText):
        """Set the result of splitting the search test.
        
            "results" is a list of strings
            "searchText" is the text on which the split was made. This is
                required for determine .currentResults{Start|End} below.
        """
        change = len(results) - len(self._rows)
        self._splits = results
        self._searchText = searchText
        self._rows = [{"split-results-tree-value": s} for s in self._splits]
        self._tree.beginUpdateBatch()
        self._tree.rowCountChanged(0, change)
        self._tree.invalidate()
        self._tree.endUpdateBatch()

    def _cacheResultSpans(self, index):
        """Determine (and cache in self._rows) the span of each split result
        up to (and including) the given "index".
        
        This is determined by working up from the start of the list of split
        results. Note that it is possible that this gets it wrong because
        the results of a regex split might not include the split-tokens and
        if any of those tokens match the split results then we can get off.
        """
        start = end = 0
        for i in range(index+1):
            if "start" in self._rows[i]:
                continue # skip already determined spans
            s = self._splits[i]
            start = self._searchText.find(s, start)
            if start == -1:
                raise ServerException(nsError.NS_ERROR_UNEXPECTED,
                    "could not find %dth split result in search text" % i)
            end = start + len(s)
            self._rows[i]["start"] = start
            self._rows[i]["end"] = end
            log.debug("split result %d: '%s': span %d-%d", i, s, start, end)

    def get_currentResultStart(self):
        index = self._tree.selection.currentIndex
        if index == -1:
            raise ServerException(nsError.NS_ERROR_FAILURE,
                                  "no currently selected split result")
        if "start" not in self._rows[index]: # lazily determine the split indices
            self._cacheResultSpans(index)
        return self._rows[index]["start"]

    def get_currentResultEnd(self):
        index = self._tree.selection.currentIndex
        if index == -1:
            raise ServerException(nsError.NS_ERROR_FAILURE,
                                  "no currently selected split result")
        if "end" not in self._rows[index]:
            self._cacheResultSpans(index)
        return self._rows[index]["end"]


    #---- nsITreeView methods
    def get_rowCount(self):
        try:
            return len(self._rows)
        except TypeError: # self._rows is None
            return 0

    def getCellText(self, row, col):
        # There is only on column, so that is ignored.
        try:
            return self._rows[row][col]
        except IndexError:
            log.error("no %sth split result" % row)
        except KeyError:
            log.error("unknown split results column: '%s'" % col)
        if type(cell) not in types.StringTypes:
            cell = str(cell)
        return cell

    def isContainer(self, row):
        return 0
    #XXX I don't think I have to implement these if isContainer == 0.
    #def isContainerEmpty(self, row):
    #    return 0
    #def isContainerOpen(self, row):
    #    return 0
    #def toggleOpenState(self, row):
    #    pass
    #def getParentIndex(self, row):
    #    pass
    #def getLevel(self, row):
    #    return 0

    def getImageSrc(self, row, col):
        #XXX Might be nice to have a icon for this.
        pass


class KoRxxRegexOptions:
    _com_interfaces_ = components.interfaces.koIRxxRegexOptions
    _reg_desc_ = "Rxx Regular Expression Options"
    _reg_clsid_ = "{2EA0D49B-EDCB-40f3-9D42-C901A3574724}"
    _reg_contractid_ = "@activestate.com/koRxxRegexOptions;1"

    def __init__(self):
        self.i = 0
        self.m = 0
        self.s = 0
        self.x = 0
        self.u = 0
        self.l = 0

    def restoreFromPrefs(self):
        prefs = components.classes["@activestate.com/koPrefService;1"].\
                getService(components.interfaces.koIPrefService).prefs
        for mod in "imsxul":
            #attr = getattr(self, mod)
            prefName = "rxx_modifier_"+mod
            #print "XXX restoring %r from %r: %r" % (attr, prefName, prefs.getBooleanPref(prefName))
            setattr(self, mod, prefs.getBooleanPref(prefName))

    def saveToPrefs(self):
        prefs = components.classes["@activestate.com/koPrefService;1"].\
                getService(components.interfaces.koIPrefService).prefs
        for mod in "imsxul":
            #attr = getattr(self, mod)
            prefName = "rxx_modifier_"+mod
            #print "XXX saving %r to %r pref" % (attr, prefName)
            prefs.setBooleanPref(prefName, getattr(self, mod))

    def repr(self, language):
        if language == "python":
            repr = []
            if self.i: repr.append('re.I')
            if self.m: repr.append('re.M')
            if self.s: repr.append('re.S')
            if self.x: repr.append('re.X')
            if self.u: repr.append('re.U')
            if self.l: repr.append('re.L')
            repr = "|".join(repr)
        elif language == "perl":
            repr = ""
            if self.i: repr += 'i'
            if self.m: repr += 'm'
            if self.d: repr += 's'
            if self.x: repr += 'x'
            #XXX Are there other Perl regex options?
            #if self.u: repr += 'u'
            #if self.l: repr += 'l'
        else:
            errmsg = "unknown regex option repr language: '%s'" % language
            #XXX koILastErrorService? log.error?
            raise ServerException(nsError.NS_ERROR_UNEXPECTED, errmsg)
        return repr


class KoRxxHandler:
    def __init__(self):
        # A cache of information on the last regex match. These should be
        # updated as appropriate after each regex match/compile.
        self._lastNumGroups = 0
        self._lastGroupNames = []
        
        from styles import StateMap
        # Mapping of style name to "style num tuples". See styles.py.
        # I don't know what it means for the tuple to have more than one
        # value. For the "Regex" language they are all 1-element tuples
        # with a style _number_.
        self.styleMap = StateMap["Regex"]

    def getLastNumGroups(self):
        return self._lastNumGroups
    def getLastGroupNames(self):
        return self._lastGroupNames

    def _style(self, target, scimoz, resultsManager):
        #XXX If lexregex.py can be made to handle it, we could start styling
        #    from scimoz.endStyled (usually >0).
        scimoz.startStyling(0, 0x1f) # styling with std 5 bits
        scimoz = scimoz
        try:
            if target == "regex":
                tokengen = lexregex.lexregex(scimoz.text,
                                             language=self.language)
            elif target == "replacement":
                tokengen = lexregex.lexreplacement(scimoz.text,
                                                   language=self.language)
            else:
                raise ValueError("unexpected style target: '%s'" % target)

            for style, index in tokengen:
                styleNum = self.styleMap[style][0]
                if not isinstance(styleNum, int):
                    styleNum = getattr(components.interfaces.ISciMoz, styleNum)
                endStyled = scimoz.endStyled
                #log.debug("style %d-%d: %r (%d)", endStyled, index,
                #          style, styleNum)
                scimoz.setStyling(index - endStyled + 1, styleNum)
        except lexregex.LexRegexError, ex:
            # There was an error in the regex -- some of the regex will
            # not have been styled. Oh well. Mark the error area if the
            # indeces of the error are noted.
            # Clear previous syntax errors.
            scimoz.startStyling(0, scimoz.INDICS_MASK)
            scimoz.setStyling(len(scimoz.text), 0)
            if ex.startIndex is not None:
                scimoz.startStyling(ex.startIndex, scimoz.INDICS_MASK)
                endIndex = ex.endIndex
                if endIndex is None:
                    endIndex = ex.startIndex
                scimoz.setStyling(endIndex - ex.startIndex + 1,
                                  scimoz.INDIC0_MASK)
            resultsManager.lexError(ex.msg)
        else:
            # Clear previous syntax errors.
            scimoz.startStyling(0, scimoz.INDICS_MASK)
            scimoz.setStyling(len(scimoz.text), 0)
            resultsManager.lexSuccess()

        # This code is called by a timer, not as the text is styled for
        # painting so if we restyled anything currently on screen, we
        # need to repaint. No "clean" way to do this (that I know of)!
        scimoz.tabWidth = scimoz.tabWidth # this forces a repaint.

    def styleRegex(self, scimoz, options, resultsManager):
        #XXX If this full styling everytime becomes a perf problem, then
        #    figure out if getting onStyleNeeded to actually work would help.
        #    This would require more of lexregex.py.
        #XXX TODO: use 'options'.
        log.debug("KoRxxHandler.styleRegex(scimoz, options, resultsManager)");
        self._style("regex", scimoz, resultsManager)

    def styleReplacement(self, scimoz, resultsManager):
        log.debug("KoRxxHandler.styleReplacment(scimoz, resultsManager)");
        self._style("replacement", scimoz, resultsManager)


class KoRxxPythonHandler(KoRxxHandler):
    _com_interfaces_ = components.interfaces.koIRxxHandler
    _reg_desc_ = "Rxx Python Handler"
    _reg_clsid_ = "{BA5DA728-DA03-4a86-8F3A-2A81791C33C9}"
    _reg_contractid_ = "@activestate.com/koRxxPythonHandler;1"
    language = "python"

    def _optionsToReFlags(self, options):
        reFlags = 0
        if options.i: reFlags |= re.I
        if options.m: reFlags |= re.M
        if options.s: reFlags |= re.S
        if options.x: reFlags |= re.X
        if options.u: reFlags |= re.U
        if options.l: reFlags |= re.L
        return reFlags

    def _compile(self, pattern, options, manager):
        """Compile the given pattern and return the Regular Expression object
        if successful, None otherwise.
        """
        reFlags = self._optionsToReFlags(options)
        regex = None
        try:
            regex = re.compile(pattern, reFlags)
        except re.error, ex:
            manager.matchError(unicode(ex))
        except IndexError, ex:
            # Trap some bugs in sre_parse.py. For example:
            #   (?P<>asdf)
            # We will rely on the lexing to report a meaningful error message.
            manager.matchError(None)
        if regex:
            self._lastNumGroups = regex.groups
            groupindex = [(num, name) for (name, num) in regex.groupindex.items()]
            groupindex.sort()
            self._lastGroupNames = [name for (num, name) in groupindex]
            return regex
        else:
            self._lastNumGroups = 0
            self._lastGroupNames = []
            return None

    def _setResults(self, manager, matchObjs, substitutions=None):
        """Set match (and possibly replacement-) results on the Rxx
        results manager.
        
            "manager" is a koIRxxResultsManager instance
            "matchObjs" is a list of Python re MatchObject's
            "substitutions" is a list of replacement strings (or None if
                not setting replacement results).
        """
        resultsView = UnwrapObject(manager.matchResultsView)
        matches = [] # a list of matches, each a list of groups
        for i in range(len(matchObjs)):
            matchObj = matchObjs[i]

            # Need to manually match up group name to group number.
            groupdict = matchObj.groupdict()
            span2name = {}
            if groupdict:
                span2name = {}
                for name in groupdict:
                    span2name[matchObj.span(name)] = name

            group = {"name": None,
                     "span": matchObj.span(0),
                     "value": matchObj.group(0)}
            if substitutions is not None:
                group["replacement"] = substitutions[i]
            groups = [group]
            for i in range(len(matchObj.groups())):
                value = matchObj.group(i+1)
                span = matchObj.span(i+1)
                group = {"name": span2name.get(span),
                         "span": span,
                         "value": value}
                if substitutions is not None:
                    group["replacement"] = ""
                groups.append(group)
            matches.append(groups)

        manager.matchSuccess(len(matches))
        resultsView.setMatches(matches)

    def _styleSearchText(self, matchObjs, scimoz):
        """Style the Search Text appropriately with the given matches."""
        #XXX It might make perf sense to move the styling of the Search Text
        #    to the Match Results Manager where clearing styling could
        #    possibly be done more efficiently. (I don't know how long
        #    clearing all styles takes.)
        scimoz.startStyling(0, 0x1f) # styling with std 5 bits
        scimoz.setStyling(len(scimoz.text), 0)
        styleNum = self.styleMap["match_highlight"][0]
        for matchObj in matchObjs:
            # Highlight the matching area.
            start, end = matchObj.span(0)
            if start != end: # perf: don't bother styling zero length match
                scimoz.startStyling(start, 0x1f) # styling with std 5 bits
                scimoz.setStyling(end-start, styleNum)
    
        # This code is called by a timer, not as the text is styled for
        # painting so if we restyled anything currently on screen, we
        # need to repaint. No "clean" way to do this (that I know of)!
        scimoz.tabWidth = scimoz.tabWidth # this forces a repaint.

    def _styleReplacedText(self, matchObjs, substitutions, scimoz):
        """Style the Replaced Text appropriately with the given replacements."""
        scimoz.startStyling(0, 0x1f) # styling with std 5 bits
        scimoz.setStyling(len(scimoz.text), 0)
        styleNum = self.styleMap["match_highlight"][0]
        offset = 0 # offset resulting from len(match) != len(substitution)
        for matchObj, substitution in zip(matchObjs, substitutions):
            # Highlight the matching area.
            start, end = matchObj.span(0)
            lenMatch = end - start
            start += offset
            offset += len(substitution) - lenMatch
            end += offset
            if start != end: # perf: don't bother styling zero length match
                scimoz.startStyling(start, 0x1f) # styling with std 5 bits
                scimoz.setStyling(end-start, styleNum)
    
        # This code is called by a timer, not as the text is styled for
        # painting so if we restyled anything currently on screen, we
        # need to repaint. No "clean" way to do this (that I know of)!
        scimoz.tabWidth = scimoz.tabWidth # this forces a repaint.

    def match(self, pattern, options, scimoz, manager):
        log.debug("KoRxxPythonHandler.match(pattern=%r, options='%s', "
                  "scimoz, manager)", pattern,
                  options and options.repr(self.language))
        searchText = scimoz.text
        matchObjs = []

        regex = self._compile(pattern, options, manager)
        if regex:
            # The regex successfully compiled, try to find a match.
            matchObj = regex.search(searchText)
            if matchObj:
                matchObjs.append(matchObj)
            if matchObjs:
                self._setResults(manager, matchObjs)
            else:
                manager.matchFailure(None)

        self._styleSearchText(matchObjs, scimoz)

    def matchAll(self, pattern, options, scimoz, manager):
        log.debug("KoRxxPythonHandler.matchAll(pattern=%r, options='%s', "
                  "scimoz, manager)", pattern,
                  options and options.repr(self.language))
        searchText = scimoz.text
        lenSearchText = len(searchText)
        matchObjs = []

        regex = self._compile(pattern, options, manager)
        if regex:
            # The regex successfully compiled, try to find matches.
            #XXX When there are many many hits (say thousands), this is
            #    disturbingly slow. It would be nice to have this be
            #    asynchronous and interruptible.
            start = 0
            while start <= lenSearchText:
                matchObj = regex.search(searchText, start)
                if matchObj:
                    matchObjs.append(matchObj)
                    if matchObj.end() - matchObj.start():
                        start = matchObj.end()
                    else: # match is zero length, advance at least one
                        start = matchObj.end() + 1
                else:
                    break
            if matchObjs:
                self._setResults(manager, matchObjs)
            else:
                manager.matchFailure(None)

        self._styleSearchText(matchObjs, scimoz)

    def split(self, pattern, options, scimoz, manager):
        log.debug("KoRxxPythonHandler.split(pattern=%r, options='%s', "
                  "scimoz, manager)", pattern,
                  options and options.repr(self.language))
        searchText = scimoz.text

        regex = self._compile(pattern, options, manager)
        if regex:
            # The regex successfully compiled, try to find a match.
            results = regex.split(searchText)
            manager.splitResultsView.setSplitResults(results, searchText)
            manager.matchSuccess(len(results))

        # clear search text highlighting
        self._styleSearchText([], scimoz)

    def replace(self, pattern, replacement, options, searchTextSciMoz,
                replacedTextSciMoz, manager):
        log.debug("KoRxxPythonHandler.replace(pattern=%r, replacement=%r, "
                  "options='%s', scimoz, scimoz, manager)", pattern,
                  replacement, options and options.repr(self.language))
        searchText = searchTextSciMoz.text
        matchObjs = []
        substitutions = []
        replacedText = searchText

        regex = self._compile(pattern, options, manager)
        if regex:
            # The regex successfully compiled, try to do a replace.
            matchObj = regex.search(searchText)
            if matchObj:
                matchObjs.append(matchObj)
                try:
                    substitution = matchObj.expand(replacement)
                    replacedText = regex.sub(replacement, searchText, 1)
                except (re.error, IndexError, ValueError), ex:
                    # e.g.:
                    #   sre_constants.error: unterminated group name
                    #   IndexError: unknown group name
                    #   ValueError: invalid literal for int(): 9
                    pass
                else:
                    substitutions.append(substitution)
            if matchObjs and substitutions:
                self._setResults(manager, matchObjs, substitutions)
            elif matchObjs:
                manager.replaceError(unicode(ex))
            else:
                manager.matchFailure(None)

        self._styleSearchText(matchObjs, searchTextSciMoz)
        replacedTextSciMoz.readOnly = 0;
        replacedTextSciMoz.text = replacedText
        replacedTextSciMoz.readOnly = 1;
        self._styleReplacedText(matchObjs, substitutions,
                                replacedTextSciMoz)

    def replaceAll(self, pattern, replacement, options, searchTextSciMoz,
                   replacedTextSciMoz, manager):
        log.debug("KoRxxPythonHandler.replaceAll(pattern=%r, replacement=%r, "
                  "options='%s', scimoz, scimoz, manager)", pattern,
                  replacement, options and options.repr(self.language))
        searchText = searchTextSciMoz.text
        lenSearchText = len(searchText)
        matchObjs = []
        substitutions = []
        replacedText = searchText

        regex = self._compile(pattern, options, manager)
        if regex:
            # The regex successfully compiled, try to do replacements.
            #XXX When there are many many hits (say thousands), this is
            #    disturbingly slow. It would be nice to have this be
            #    asynchronous and interruptible.
            start = 0
            while start <= lenSearchText:
                matchObj = regex.search(searchText, start)
                if matchObj:
                    matchObjs.append(matchObj)
                    try:
                        substitution = matchObj.expand(replacement)
                    except (re.error, IndexError, ValueError), ex:
                        # e.g.:
                        #   sre_constants.error: unterminated group name
                        #   IndexError: unknown group name
                        #   ValueError: invalid literal for int(): 9
                        pass
                    else:
                        substitutions.append(substitution)

                    if matchObj.end() - matchObj.start():
                        start = matchObj.end()
                    else: # match is zero length, advance at least one
                        start = matchObj.end() + 1
                else:
                    break

            try:
                replacedText = regex.sub(replacement, searchText)
            except (re.error, IndexError, ValueError), ex:
                # e.g.:
                #   sre_constants.error: unterminated group name
                #   IndexError: unknown group name
                #   ValueError: invalid literal for int(): 9
                pass

            if matchObjs and substitutions:
                self._setResults(manager, matchObjs, substitutions)
            elif matchObjs:
                manager.replaceError(unicode(ex))
            else:
                manager.matchFailure(None)

        self._styleSearchText(matchObjs, searchTextSciMoz)
        replacedTextSciMoz.readOnly = 0;
        replacedTextSciMoz.text = replacedText
        replacedTextSciMoz.readOnly = 1;
        self._styleReplacedText(matchObjs, substitutions,
                                replacedTextSciMoz)
