# Copyright (c) 2000-2001 ActiveState Corporation.
# See the file LICENSE.txt for licensing information.

from xpcom import components
import string
import re
import logging


log = logging.getLogger("koScintillaController")

command_map = {
    'cmd_selectHome' : 'vCHomeWrapExtend',
    'cmd_selectEnd' : 'lineEndWrapExtend',
    'cmd_selectCharPrevious' : 'charLeftExtend',
    'cmd_selectCharNext' : 'charRightExtend',
    'cmd_pageUp' : 'pageUp',
    'cmd_pageDown' : 'pageDown',
    'cmd_selectPageUp' : 'pageUpExtend',
    'cmd_selectPageDown' : 'pageDownExtend',
    'cmd_selectLinePrevious' : 'lineUpExtend',
    'cmd_selectLineNext' : 'lineDownExtend',
    'cmd_undo' : 'undo',
    'cmd_redo' : 'redo',
    'cmd_cut' : 'cut',
    'cmd_copy' : 'copy',
    'cmd_paste' : 'paste',
    'cmd_selectWordLeft' : 'wordLeftExtend',
    'cmd_selectWordRight' : 'wordRightExtend',
    'cmd_wordPartLeftExtend' : 'wordPartLeftExtend',
    'cmd_wordPartRightExtend' : 'wordPartRightExtend',
    'cmd_deleteWordLeft' : 'delWordLeft',
    'cmd_deleteWordRight' : 'delWordRight',
    'cmd_documentHome' : 'documentStart',
    'cmd_documentEnd' : 'documentEnd',
    'cmd_selectDocumentHome' : 'documentStartExtend',
    'cmd_selectDocumentEnd' : 'documentEndExtend',
    'cmd_editSelectAll' : 'selectAll',
    'cmd_delete' : 'clear',
    'cmd_back' : 'deleteBack',
    'cmd_lineScrollUp' : 'lineScrollUp',
    'cmd_lineScrollDown' : 'lineScrollDown',
    'cmd_lineCut' : 'lineCut',
    'cmd_lineDelete' : 'lineDelete',
#    'cmd_lineTranspose' : 'lineTranspose',  # commented out because its undo behavior is wrong.
    'cmd_fontZoomIn' : 'zoomIn',
    'cmd_fontZoomOut' : 'zoomOut',
    'cmd_toggleOvertype' : 'editToggleOvertype',
    #'cmd_newline' : 'newLine',
}

undoable_commands = {
    'cmd_lineCut' : 1,
    'cmd_lineDelete' : 1,
    'cmd_lineTranspose' : 1,
    'cmd_deleteWordLeft' : 1,
    'cmd_deleteWordRight' : 1,
}

class koScintillaController:
    _com_interfaces_ = components.interfaces.ISciMozController
    _reg_clsid_ = "{726cc885-6d17-48af-b8a6-c9b759f1fe6b}"
    _reg_contractid_ = "@ActiveState.com/scintilla/controller;1"

    def init(self, scimoz):
        self.scimoz = scimoz
        self._lastcutposition = None
        # A cached collection of boolean attributes.
        # No key == "dunno if boolean"
        # Key == None == definitely not bool.
        # Else - key == scimoz attribute name.
        self.bool_attributes = {}

    def test_scimoz(self, scimoz):
        self.init(scimoz)
        ScintillaControllerTestCase.controller = self
        testCases = [ScintillaControllerTestCase]
        sciutils.runSciMozTests(testCases, scimoz)

    def isCommandEnabled( self, command_name ):
        # Result: boolean
        # In: param0: wstring
        meth = getattr(self, "_is_%s_enabled" % (str(command_name),), None)
        if meth is None:
            # Handle the fact it may be a bool property that can be toggled.
            # (and cache the fact, to speed things up 2nd and later times)
            attr = self.bool_attributes.get(command_name, 0)
            if attr == 0: # not in map
                attr = prop_name = command_map.get(command_name)
                if attr is not None:
                    attr = getattr(self.scimoz, attr)
                    if type(attr) == type(0):
                        # Cache for next time.
                        self.bool_attributes[command_name] = prop_name
                        # And to avoid re-fetching, get out here!
                        return attr
                    # Not an attribute we can use - that is OK.
                    attr = None

            if attr is None:
                # No custom function, and no scintilla integer property of that name.
                # Assume OK.
                rc = 1
            else:
                # We have a remembered property name - use it.
                rc = getattr(self.scimoz, attr)
        else:
            # Custom function - do it
            rc = meth()
        return rc

    def supportsCommand( self, command_name ):
        # Result: boolean
        # In: param0: wstring

        return command_name in command_map or hasattr(self, '_do_'+command_name)

    def doCommand( self, command_name ):
        # Result: void - None
        # In: param0: wstring
        sm = self.scimoz
        old_sel_exists = sm.selectionEnd != sm.selectionStart
        currentPos = sm.currentPos
        if command_name == 'cmd_undo':
            if sm.autoCActive():
                sm.autoCCancel()
            elif sm.callTipActive():
                sm.callTipCancel()
            targetPos = None
            if self._lastcutposition is not None and sm.currentPos == self._lastcutposition:
                # we need to remove a line from the undo buffer
                lines = self._cutbuffer.splitlines(1)
                self._cutbuffer = ''.join(lines[:-1])
                sm.copyText(len(self._cutbuffer), self._cutbuffer)
                # note where we want to be post undo
                targetPos = self._lastcutposition
            sm.undo()
            if targetPos:
                sm.currentPos = sm.anchor = targetPos
            return
        if command_name == 'cmd_copy':
            # If there is a selection, we just do the usual cut
            if sm.selectionStart != sm.selectionEnd:
                sm.copy()
                sm.sendUpdateCommands("select")
                sm.sendUpdateCommands("clipboard")
                return
            # if there's no selection, we copy the current line, being careful to leave
            # the cursor in its original position.
            # but should we?  Most of the time one will want to place the line somewhere else.
            # Hmm -- I think I'll leave the cursor at the beginning of the copied line for now.
            oldCurrentPos = sm.currentPos
            lineStart = sm.lineFromPosition(sm.currentPos)
            sm.selectionStart = sm.positionFromLine(lineStart)
            sm.selectionEnd = sm.positionFromLine(lineStart+1)
            sm.copy()
            sm.sendUpdateCommands("select")
            sm.sendUpdateCommands("clipboard")
            sm.currentPos = sm.selectionEnd = sm.selectionStart
            sm.chooseCaretX()
            self._lastcutposition = None 
        elif command_name == 'cmd_cut':
            # If there is a selection, we just do the usual cut
            if sm.selectionStart != sm.selectionEnd:
                sm.cut()
                self._lastcutposition = None
                sm.sendUpdateCommands("clipboard")
                return
            # Do nothing at end of file except if there's stuff to the left
            if sm.currentPos == sm.textLength and \
               sm.getColumn(sm.currentPos) == 0: return 
            # if there's no selection, we get to do our fancy cut.
            # If the last cut position was different that the current position, this is the 'first cut' -- e.g. a lineCut
            # We're cutting a line -- either the first, or possibly a subsequent one
            lineNo = sm.lineFromPosition(sm.currentPos)
            lineStart = sm.positionFromLine(lineNo)
            lineEnd = min(sm.positionFromLine(lineNo+1), sm.textLength)
            self._doSmartCut(lineStart, lineEnd)
            return
        methname= '_do_'+command_name
        attr = getattr(self, methname, None)
        if attr is None:
            attr = getattr(sm, command_map[command_name])
        # If we fetch an attribute and it is a number..
        if type(attr)==type(0): # Assume boolean
            setattr(sm, command_map[command_name], not attr)
        # Usually it will be a method..
        elif callable(attr):
            attr()
        # and hopefully this will never happen!
        else:
            raise TypeError, "The command map entry '%s' yielded a '%r' - dunno what to do with it" % (command_name, attr)
        # We only need to send a command update when the selection changes _shape_, not location
        # (ie, only when it changes from no selection to selection, or vice-versa.
        if command_name not in ('cmd_redo', 'cmd_killLine'):
            # We've done something else, that means that we're not in a multiple-ctrl-x mode anymore
            self._lastcutposition = None 
        new_sel_exists = sm.selectionEnd != sm.selectionStart
        if old_sel_exists != new_sel_exists:
            sm.sendUpdateCommands("select")

    def _doSmartCut(self, start, end):
        sm = self.scimoz
        sm.targetStart = start
        sm.targetEnd = end
        line = sm.getTextRange(start, end)
        sm.beginUndoAction()
        try:
            sm.replaceTarget(0, '')
            if self._lastcutposition != sm.currentPos:
                # first cut -- lineCut
                self._cutbuffer = line
                self._lastcutposition = sm.currentPos
            else:
                # If we're here, it's because someone did Cut twice w/o a
                # selection -- we need to accumulate the lines into the cut
                # buffer
                self._cutbuffer += line
            # Add one, otherwise a LF on a Windows CRLF line terminator
            # does not make it into the clipboard. Note that we *should*
            # add one on Unix as well even though there the EOL is only
            # one character. (c.f. bug 29236).
            sm.copyText(len(self._cutbuffer)+1, self._cutbuffer)
            sm.chooseCaretX()
        finally:
            sm.endUndoAction()
        sm.sendUpdateCommands("clipboard")

    def onEvent( self, param0 ):
        # Result: void - None
        # In: param0: wstring
        log.warn("Scintilla controller caught an event: ", param0)

    def _has_sel(self):
        return self.scimoz.selectionEnd != self.scimoz.selectionStart
    # Specific command handlers
    def _is_cmd_undo_enabled(self):
        return self.scimoz.canUndo()
    def _is_cmd_redo_enabled(self):
        return self.scimoz.canRedo()
    def _is_cmd_paste_enabled(self):
        return self.scimoz.canPaste()
    def _is_cmd_cut_enabled(self):
        return not self.scimoz.readOnly
    def _is_cmd_copy_enabled(self):
        return 1
    def _is_cmd_delete_enabled(self):
        return 1

    def _is_cmd_pasteAndSelect_enabled(self):
        return self.scimoz.canPaste()
    def _do_cmd_pasteAndSelect(self):
        start = self.scimoz.currentPos
        self.scimoz.paste()
        self.scimoz.anchor = start

    def _do_cmd_endOfWord(self):
        sm = self.scimoz
        if sm.currentPos == sm.textLength:
            return
        sm.beginUndoAction()
        try:
            curChar = sm.getWCharAt(sm.currentPos)
            nextChar = sm.getWCharAt(sm.positionAfter(sm.currentPos))
            if charClass.get(curChar, 'special') != charClass.get(nextChar, 'special'):
                sm.charRight()
            while sm.currentPos < sm.textLength and charClass.get(curChar, 'special') == charClass.get(nextChar, 'special'):
                sm.charRight()
                curChar = nextChar
                nextChar = sm.getWCharAt(sm.currentPos)
        finally:
            sm.endUndoAction()
        return 1

    def _do_cmd_endOfWordExtend(self):
        sm = self.scimoz
        if sm.currentPos == sm.textLength:
            return
        sm.beginUndoAction()
        try:
            curChar = sm.getWCharAt(sm.currentPos)
            nextChar = sm.getWCharAt(sm.positionAfter(sm.currentPos))
            if charClass.get(curChar, 'special') != charClass.get(nextChar, 'special'):
                sm.charRightExtend()
            while sm.currentPos < sm.textLength and charClass.get(curChar, 'special') == charClass.get(nextChar, 'special'):
                sm.charRightExtend()
                curChar = nextChar
                nextChar = sm.getWCharAt(sm.currentPos)
        finally:
            sm.endUndoAction()
        return 1

    def _do_cmd_beginningOfWord(self):
        sm = self.scimoz
        if sm.currentPos == 0:
            return
        sm.beginUndoAction()
        try:
            curChar = sm.getWCharAt(sm.currentPos)
            nextChar = sm.getWCharAt(sm.positionBefore(sm.currentPos))
            if charClass.get(curChar, 'special') != charClass.get(nextChar, 'special'):
                sm.charLeft()
            curChar = sm.getWCharAt(sm.currentPos)
            nextChar = sm.getWCharAt(sm.positionBefore(sm.currentPos))
            while sm.currentPos and charClass.get(curChar, 'special') == charClass.get(nextChar, 'special'):
                sm.charLeft()
                curChar = nextChar
                nextChar = sm.getWCharAt(sm.positionBefore(sm.currentPos))
        finally:
            sm.endUndoAction()
        return 1

    def _do_cmd_beginningOfWordExtend(self):
        sm = self.scimoz
        if sm.currentPos == 0:
            return
        sm.beginUndoAction()
        try:
            curChar = sm.getWCharAt(sm.currentPos)
            nextChar = sm.getWCharAt(sm.positionBefore(sm.currentPos))
            if charClass.get(curChar, 'special') != charClass.get(nextChar, 'special'):
                sm.charLeftExtend()
            curChar = sm.getWCharAt(sm.currentPos)
            nextChar = sm.getWCharAt(sm.positionBefore(sm.currentPos))
            while sm.currentPos and charClass.get(curChar, 'special') == charClass.get(nextChar, 'special'):
                sm.charLeftExtend()
                curChar = nextChar
                nextChar = sm.getWCharAt(sm.positionBefore(sm.currentPos))
        finally:
            sm.endUndoAction()
        return 1

    def _do_cmd_lineTranspose(self):
        sm = self.scimoz
        sm.beginUndoAction()
        try:
            sm.lineTranspose();
        finally:
            sm.endUndoAction()
        return 1
    
    def _do_cmd_join(self):
        sm = self.scimoz
        sm.beginUndoAction()
        try:
            lineNo = sm.lineFromPosition(sm.currentPos)
            lineStart = sm.positionFromLine(lineNo)
            lineEnd = sm.getLineEndPosition(lineNo)
            whitespaceToLeft = False
            if sm.getTextRange(sm.currentPos, lineEnd).strip():
                # We're in the middle of a line -- assume we're at the
                # end of the line.
                sm.lineEnd()
            if sm.getWCharAt(sm.currentPos-1) == ' ':
                whitespaceToLeft = True
            sm.lineEnd()
            # delete all but last whitespace characters to the left
            # of cursor
            start = end = sm.currentPos
            endBuffer = sm.textLength
            lineStart = sm.positionFromLine(sm.lineFromPosition(sm.currentPos))
            # if there is non-whitespace to the left, then delete any
            # immediately preceding whitespace
            stuffToLeft = sm.getTextRange(lineStart, sm.currentPos)
            if stuffToLeft.strip():
                while start >= lineStart+1 and sm.getWCharAt(start-1) == ' ':
                    start -= 1
                replacement = ' '
            else:
                replacement = ''
            while end <= endBuffer and sm.getWCharAt(end) in ' \r\n':
                end += 1
            sm.targetStart = start
            sm.targetEnd = end
            sm.replaceTarget(1, replacement)
            if whitespaceToLeft:
                sm.gotoPos(sm.targetEnd)
            else:
                sm.gotoPos(sm.targetStart)
        finally:
            sm.endUndoAction()

    def _do_cmd_linePrevious(self):
        sm = self.scimoz
        sm.currentPos = min(sm.currentPos, sm.anchor)
        sm.lineUp()
        
    def _do_cmd_lineNext(self):
        sm = self.scimoz
        sm.currentPos = max(sm.currentPos, sm.anchor)
        sm.lineDown()
        
    def _do_cmd_left(self):
        sm = self.scimoz
        if sm.currentPos != sm.anchor:
            sm.currentPos = sm.anchor = min(sm.currentPos, sm.anchor)
        else:
            sm.currentPos = min(sm.currentPos, sm.anchor)
            sm.charLeft()
        
    def _do_cmd_right(self):
        sm = self.scimoz
        if sm.currentPos != sm.anchor:
            sm.currentPos = sm.anchor = max(sm.currentPos, sm.anchor)
        else:
            sm.currentPos = max(sm.currentPos, sm.anchor)
            sm.charRight()

    def _do_cmd_wordLeft(self):
        sm = self.scimoz
        sm.currentPos = min(sm.currentPos, sm.anchor)
        sm.wordLeft()
        
    def _do_cmd_wordRight(self):
        sm = self.scimoz
        sm.currentPos = max(sm.currentPos, sm.anchor)
        sm.wordRight()
        
    def _do_cmd_wordPartLeft(self):
        sm = self.scimoz
        sm.currentPos = min(sm.currentPos, sm.anchor)
        sm.wordPartLeft()
        
    def _do_cmd_wordPartRight(self):
        sm = self.scimoz
        sm.currentPos = max(sm.currentPos, sm.anchor)
        sm.wordPartRight()

    def _do_cmd_home(self):
        sm = self.scimoz
        sm.currentPos = min(sm.currentPos, sm.anchor)
        sm.vCHomeWrap()  # XXX should make this a pref.
        
    def _do_cmd_end(self):
        sm = self.scimoz
        sm.currentPos = max(sm.currentPos, sm.anchor)
        sm.lineEndWrap()

    def _is_cmd_transpose_enabled(self):
        return 1 # not really, but we'll deal with edge cases below

    def _do_cmd_transpose(self):
        # transpose two characters to the left
        if self.scimoz.getColumn(self.scimoz.currentPos) < 2:
            return
        
        # this particular sequence (others are possible) has good Undo behavior
        self.scimoz.beginUndoAction()
        try:
            self.scimoz.charLeftExtend()
            self.scimoz.cut()
            self.scimoz.charLeft()
            self.scimoz.paste()
            self.scimoz.charRight()
            #sm.charRight() # this one is required for Emacs compatibility, but I don't see the use case.
        finally:
            self.scimoz.endUndoAction()

    def _do_cmd_transposeWords(self):
        # transpose two words to the left (or current word and previous word if in a word)
        # this particular sequence (others are possible) has good Undo behavior
        sm = self.scimoz
        sm.beginUndoAction()
        try:
            sm.wordLeft()
            sm.wordLeftExtend()
            sm.cut()
            sm.wordLeft()
            sm.paste()
            sm.wordRight()
        finally:
            sm.endUndoAction()
            
    def _do_cmd_killLine(self):
        # emacs-style 'kill': if there is nothing but whitespace on the line,
        # that's the same as a cut. If there is something other than whitespace,
        # then it's equivalent to "cut the current line _not including_ the EOL.
        sm = self.scimoz
        lineNo = sm.lineFromPosition(sm.currentPos)
        endLine = sm.getLineEndPosition(lineNo)
        line = sm.getTextRange(sm.currentPos, endLine)
        if line.strip():
            end = endLine
        else:
            end = min(sm.positionFromLine(lineNo+1), sm.textLength)
        self._doSmartCut(sm.currentPos, end)

charClass = {}
for x in string.letters + string.digits + '_':
    charClass[x] = 'alpha'
for x in string.whitespace:
    charClass[x] = 'whitespace'



import sciutils
class ScintillaControllerTestCase(sciutils.SciMozTestCase):
    controller = None
    
    def assertCmdResultIs(self, buffer, cmdName, expected, language="Python"):
        self._setupSciMoz(buffer, language)
        cmd = getattr(self.controller, "_do_cmd_%s" % cmdName)
        cmd()
        eText, eCurrentPos = self._parseBuffer(expected)
        text, currentPos = self.scimoz.text, self.scimoz.currentPos
        got = text[:currentPos] + "<|>" + text[currentPos:]
        comparison = """
--- Expected: -----------------------------------------
%s
--- Got: ----------------------------------------------
%s
-------------------------------------------------------
""" % (expected, got)
        self.assertEqual(text, eText,
                         ("unexpected text after cmd '%s':"%cmdName)
                         + comparison)
        self.assertEqual(currentPos, eCurrentPos,
                         ("unexpected currentPos after cmd '%s':"%cmdName)
                         + comparison)

    def test_cmd_join(self):
        self.assertCmdResultIs("foo\n<|>bar\nspam", "join",
                               "foo\nbar<|> spam")


