#!python
# Copyright (c) 2004 ActiveState Corp, a division of Sophos Plc.
# See the file LICENSE.txt for licensing information.

"""Code Intel Type Determination Lang (CITDL, "citadel") processing."""

from __future__ import generators

import os
import sys
import time
import md5
import pprint

import logging
import process

from codeintel.common import *  # ScanRequest and exceptions are defined here


#---- global variables

log = logging.getLogger("citdl")
#log.setLevel(logging.INFO)

# CIDB table row indeces.
(S_ID, S_FILE_ID, S_SCAN_ID, S_MODULE_ID, S_SCOPE_ID, S_SCOPE, S_NAME,
 S_SIGNATURE, S_LINE, S_LINEEND, S_TYPE, S_ATTRIBUTES, S_DOC) = range(13)
(M_ID, M_FILE_ID, M_SCAN_ID, M_NAME, M_LINE, M_ATTRIBUTES, M_DOC) = range(7)
(I_ID, I_FILE_ID, I_SCAN_ID, I_MODULE_ID, I_SCOPE_ID, I_LINE, I_MODULE,
 I_SYMBOL, I_ALIAS) = range(9)


#---- Import handlers (import semantics vary among languages)

def _getFileId(cu, path):
    cpath = canonicalizePath(path)
    cu.execute("SELECT id FROM file WHERE compare_path=%s", cpath)
    if not cu.rowcount:
        raise CodeIntelError("could not find file in CIDB: '%s'" % path)
    else:
        return int(cu.fetchone()[0])
        
def getImportHandler(language):
    lang2handler = {
        "Python": PythonImportHandler,
        "Perl": PerlImportHandler,
        "PHP": PHPImportHandler,
        "JavaScript": JavaScriptImportHandler,
        "Tcl": TclImportHandler,
    }
    if language not in lang2handler:
        raise CodeIntelError("cannot handle '%s' imports: there is no "
                             "registered ImportHandler class for "
                             "language '%s'" % (language, language))
    return lang2handler[language]()

class ImportHandler:
    """Virtual base class for language-specific "import"-statement handlers.
    
    The basic job of an import handler is to convert an import statement (i.e.
    a row in the 'import' table) into a row in the CIDB 'module' table. Doing
    this depends on language-specific import semantics.
    
    A fundamental part of import resolution is the search path. Here the
    search path is broken into three parts:
        - "core" path: built-in to the interpreter/compiler, generally
          dependent on the installation location.
        - "env" path: additional directories specified in a special
          environment variable, e.g. PYTHONPATH, PERL5LIB
        - "custom" path: additional "out-of-band" directories
    
    Each language-specific ImportHandler is a singleton, as controlled by
    Manager.getImportHandler().
    """
    PATH_ENV_VAR = None
    corePath = None
    envPath = None
    customPath = None

    def setCustomPath(self, path):
        """Specify some custom search directories."""
        self.customPath = path
    def setEnvPath(self, value=None):
        """Specify the value of the PATH-style environment variable.
        
            "value" is the appropriate environment variable value, e.g.:
                "C:\trentm\mylib;C:\shared\lib-python". If value is None then
                the value will be retrieved from os.environ.

        This will lazily be called if necessary.
        """
        path = []
        if value is None:
            if self.PATH_ENV_VAR:
                path = os.environ.get(self.PATH_ENV_VAR, "").split(os.pathsep)
        else:
            path = value.split(os.pathsep)
        self.envPath = path
    def setCorePath(self, compiler=None, extra=None):
        """Specify the data needed to determine the core search path.
        
            "compiler" is the path to the language compiler/interpreter.
                If not specified then the first appropriate
                compiler/interpreter on the path is used.
            "extra" (optional) can be used to specify required extra
                data for determining the core import path.  For example,
                the PHP-specific implementation uses this to specify the
                "php.ini"-config-file path.

        This will lazily be called if necessary.
        """
        # Sub-classes must implement this and set self.corePath as a
        # result (to [] if it could not be determined).
        raise NotImplementedError("setCorePath: pure virtual method call")

    def _getPath(self, cwd=None):
        """Put all the path pieces together and return that list.
        
        If "cwd" is specified, it is prepended to the list. (In many languages
        the directory of the file with the import statement is first on the
        module search path.)
        """
        if self.corePath is None: self.setCorePath()
        if self.envPath is None: self.setEnvPath()
        if cwd is not None:
            path = [cwd]
        else:
            path = []
        if self.customPath:
            path += self.customPath
        path += self.envPath
        path += self.corePath
        return path

    def findModuleOnDisk(self, module, submodule, cwd):
        """Find the full path to the given module on disk.
        
            "module" is a the module name (_can_ include dots, e.g. xml.sax)
            "submodule" is a possible sub-module to also include in the
                import path.
            "cwd" is the directory of the module containing the import stmt.
        
        If successful, returns a 3-tuple:
            (<module path>, "module"|"submodule", <scannable>,
        where,
            the second element indicates if the returned module row
                corresponds to the given "module" or "submodule" argument.
            <scannable> is a boolean indicating if this file is likely
                scannable.

        Otherwise, raises an ImportError.
        """
        raise NotImplementedError("findModuleOnDisk: pure virtual method call")

    def findModule(self, cu, module, submodule, cwd=None):
        """Find the 'module' table row corresponding to the given import row.
        
            "cu" is an open CIDB cursor
            "module" is the module name (or path, e.g. "xml.sax")
            "submodule" is a possible sub-module to "name" in an import
                statement, some languages allow you to separate the two to
                control the imported symbol name. E.g., In the following
                Python statement "bar" might be a submodule of the package
                "foo", or it might just be a symbol of module "foo":
                    from foo import bar
            "cwd" is the directory of the module containing the import stmt.

        If successful, returns a 2-tuple:
            (<module row>, "module"|"submodule")
        The second element indicates if the returned module row corresponds
        to the given "module" or "submodule" argument.
        
        Algorithm:
        - Look for the appropriate file on disk, using the import path.
        - If found, see if we have this file in the CIDB.
            - If it's in the CIDB, then find the module with the appropriate
              name from that file.
        - If haven't found a matching module row yet, then lookup the full
          import path (and language) in the 'module' table and get all hits.
          This has the potential to "hit" modules that aren't in the search
          path, but a false positive is better than nothing in this case.
        - If still no hits, then raise an error that describes the problem
          in as helpful a way as possible for the user.
        
        XXX To re-scan or not to re-scan:
            May want to schedule a re-scan of the file whether it is in the
            CIDB or not. Or maybe just if it isn't. Or maybe this is up to
            the CIX importing to kick off and we don't need to worry about it.
            In any case, if we _do_ re-scan files already in the CIDB then
            we might want to exclude files whose "md5" value is marked with
            the special "*", meaning: "don't re-scan".
        """
        raise NotImplementedError("findModule: pure virtual method call")

    def genScannableFiles(self, path=None, skipRareImports=False,
                          importableOnly=False):
        """Generate scannable files on the import path.
        
            "path" (optional) is an import path to load. If not specified
                the default import path is used.
            "skipRareImports" (optional, default false) is a boolean
                indicating if files unlikely to be imported/searched-for/used
                should be skipped. This can be specified to speed up, for
                example, scanning *all* files during a batch update of a
                language installation.
            "importableOnly" (optional, default false) is a boolean
                indicating if only those files that are importable from
                the given path should be included. For example a Python
                file in a subdirectory cannot be imported if that dir
                does not have a package-defining "__init__.py" file.
        """
        raise NotImplementedError("genScannableFiles: pure virtual method call")



class PythonImportHandler(ImportHandler):
    PATH_ENV_VAR = "PYTHONPATH"

    def __init__(self):
        self.__stdCIXScanId = None

    def _shellOutForPath(self, compiler):
        import process
        argv = [compiler, "-c", "import sys; print '\\n'.join(sys.path)"]
        # Can't use -E to ignore PYTHONPATH because older versions of
        # Python don't have it (e.g. v1.5.2).
        env = dict(os.environ)
        if "PYTHONPATH" in env: del env["PYTHONPATH"]
        if "PYTHONHOME" in env: del env["PYTHONHOME"]
        if "PYTHONSTARTUP" in env: del env["PYTHONSTARTUP"]

        p = process.ProcessOpen(argv, env=env)
        retval = p.wait()
        output = p.stdout.read()
        p.close()
        path = [line for line in output.splitlines(0)]
        if path and (path[0] == "" or path[0] == os.getcwd()):
            del path[0] # cwd handled separately
        return path

    def setCorePath(self, compiler=None, extra=None):
        if compiler is None:
            import which
            assert which._version_ >= (1, 0, 0)
            compiler = which.which("python")
        self.corePath = self._shellOutForPath(compiler)

    def _getStdCIXScanId(self, cu):
        if not self.__stdCIXScanId:
            #start = time.clock() #XXX
            cu.execute("SELECT scan.id FROM scan, file, language "
                       " WHERE language.name='Python'"
                       "   AND language.id=file.language_id"
                       "   AND scan.file_id=file.id"
                       "   AND scan.generator='StdCIX'")
            #end = time.clock() #XXX
            #print "XXX _getStdCIXScanId: %.3fs" % (end-start)
            if cu.rowcount:
                self.__stdCIXScanId = int(cu.fetchone()[0])
        return self.__stdCIXScanId

    def findModule(self, cu, module, submodule, cwd=None):
        modrow = kind = None

        # Some specific modules are much better defined in 'python.cix',
        # if it has been imported, than a scan of the associated .py file.
        #XXX The _correct_ algorithm would probably be to (1) find the
        #    proper file to import via findModuleOnDisk and (2) if that
        #    file was from the stdlib that the overrides represent, only
        #    _then_ would the override take effect. Currently this will
        #   override a user's own os.py. This is not deemed a big deal.
        overrides = ["os", "os.path", "re"]
        if module in overrides:
            stdCIXScanId = self._getStdCIXScanId(cu)
            if stdCIXScanId:
                sql = "SELECT * FROM module WHERE scan_id=%d AND name=%%s"\
                      % stdCIXScanId
                if submodule:
                    cu.execute(sql, module+'.'+submodule)
                    if cu.rowcount:
                        modrow = cu.fetchone()
                        kind = "submodule"
                if not modrow:
                    cu.execute(sql, module)
                    if cu.rowcount:
                        modrow = cu.fetchone()
                        kind = "module"
        
        if not modrow:
            # Look for the appropriate file on disk, using the import path.
            try:
                modfile, kind, scannable = self.findModuleOnDisk(module,
                                                                 submodule,
                                                                 cwd)
            except ImportError, ex:
                modfile = kind = None
               
            # If found, see if we have this file and module in the CIDB.
            file_id = None
            if modfile:
                try:
                    file_id = _getFileId(cu, modfile)
                except CodeIntelError, ex:
                    pass
                if file_id:
                    if kind == "module":
                        modname = module.split('.')[-1]
                    elif kind == "submodule":
                        modname = submodule
                    cu.execute("SELECT * FROM module WHERE file_id=%d AND name=%s",
                               file_id, modname)
                    if cu.rowcount:
                        modrow = cu.fetchone()

        # If haven't found a matching module row yet, then lookup the full
        # import path (and language) in the 'module' table and get all hits.
        if not modrow:
            #XXX Might want database index specifically for this?
            sql = "SELECT module.* FROM module, file, language "\
                  " WHERE language.name='Python' "\
                  "   AND language.id=file.language_id "\
                  "   AND module.file_id=file.id "\
                  "   AND module.name=%s"
            if submodule:
                cu.execute(sql, module+'.'+submodule)
                if cu.rowcount:
                    modrow = cu.fetchone()
                    kind = "submodule"
            if not modrow:
                cu.execute(sql, module)
                if cu.rowcount:
                    modrow = cu.fetchone()
                    kind = "module"
            if cu.rowcount > 1:
                log.info("found more than one module table row for module "
                         "'%s' (submodule '%s'): first one wins",
                         module, submodule)

        if not modrow:
            if not submodule:
                desc = "'%s' module" % module
            else:
                desc = "'%s.%s' or '%s' module" % (module, submodule, module)
            raise CodeIntelError("couldn't find %s in database" % desc)
        else:
            return (modrow, kind)

    def findModuleOnDisk(self, module, submodule, cwd=None):
        r"""        
            >>> h = PythonImportHandler()
            >>> h.findModuleOnDisk("cmd", None)
            ('C:\\Python22\\lib\\cmd.py', 'module', 1)
            >>> h.findModuleOnDisk("xml", None)
            ('C:\\Python22\\lib\\xml\\__init__.py', 'module', 1)
            >>> h.findModuleOnDisk("xml", "sax")
            ('C:\\Python22\\lib\\xml\\sax\\__init__.py', 'submodule', 1)
            >>> h.findModuleOnDisk("xml", "*")
            ('C:\\Python22\\lib\\xml\\__init__.py', 'module', 1)
            >>> h.findModuleOnDisk("xml", "__version__")
            ('C:\\Python22\\lib\\xml\\__init__.py', 'module', 1)
            >>> h.findModuleOnDisk("xml.sax", None)
            ('C:\\Python22\\lib\\xml\\sax\\__init__.py', 'module', 1)
            >>> h.findModuleOnDisk("xml.sax", "handler")
            ('C:\\Python22\\lib\\xml\\sax\\handler.py', 'submodule', 1)
            >>> h.findModuleOnDisk("xml.sax", "parse")
            ('C:\\Python22\\lib\\xml\\sax\\__init__.py', 'module', 1)
            >>> h.findModuleOnDisk("xml.sax", "*")
            ('C:\\Python22\\lib\\xml\\sax\\__init__.py', 'module', 1)
            >>> h.findModuleOnDisk("_tkinter", "*")
            ('C:\\Python22\\DLLs\\_tkinter.pyd', 'module', 0)
        """
        if isinstance(cwd, unicode):
            try:
                cwd = str(cwd)  # imp.find_module() doesn't like unicode paths
            except:
                log.warn("could not convert unicode CWD to string for Python"
                         "module lookup: lookup might fail: cwd=%r" % cwd)
                cwd = None
        path = self._getPath(cwd)

        import imp
        modparts = module.split('.')
        modpath = description = None
        for i in range(len(modparts)):
            modpart = modparts[i]
            fin, modpath, desc = imp.find_module(modpart, path)
            if fin: fin.close()
            path = [modpath]
        kind = "module"
        
        # Check to see if the given sub-module can be tacked on.
        if submodule:
            try:
                sfin, smodpath, sdesc = imp.find_module(submodule, [modpath])
                if sfin: sfin.close()
            except ImportError:
                pass
            else:
                kind = "submodule"
                modpath = smodpath
                desc = sdesc

        # If the "module path" is a package directory, then convert it to
        # the appropriate _file_.
        if desc[2] == imp.PKG_DIRECTORY:
            fin, modpath, desc = imp.find_module("__init__", [modpath])

        # The import is scannable if it is a plain Python source file.
        suffix, mode, type = desc
        scannable = type == imp.PY_SOURCE

        return (modpath, kind, scannable)

    def _findScannableFiles(self,
                            (files, searchedDirs, skipRareImports,
                             importableOnly),
                            dirname, names):
        if sys.platform.startswith("win"):
            cpath = dirname.lower()
        else:
            cpath = dirname
        if cpath in searchedDirs:
            while names:
                del names[0]
            return
        else:
            searchedDirs[cpath] = 1
        if skipRareImports:
            if (os.path.basename(dirname) == "encodings"
                and "undefined.py" in names):
                # Skip most of the specific encoding definitions (saves
                # about 50 files).
                names = [n for n in names if n == "__init__.py"
                         or os.path.splitext(n)[0].endswith("_codec")]
        for i in range(len(names)-1, -1, -1): # backward so can del from list
            path = os.path.join(dirname, names[i])
            if os.path.isdir(path):
                if skipRareImports:
                    # Skip Python's test package (saves over 200 files)
                    # and other likely test dirs.
                    if names[i] in ("test", "tests"):
                        del names[i]
                        continue
                if importableOnly:
                    possibles = [os.path.join(path, "__init__.py"),
                                 os.path.join(path, "__init__.pyc"),
                                 os.path.join(path, "__init__.pyo")]
                    for possible in possibles:
                        if os.path.isfile(possible):
                            break
                    else:
                        del names[i] # don't traverse non-package dirs
                        continue
                if path.endswith(os.path.join("win32com", "gen_py")):
                    del names[i]
                    continue
            elif os.path.splitext(names[i])[1] in (".py", ".pyw"):
                #XXX The list of Python extensions should be settable on
                #    the ImportHandler and Komodo should set whatever is
                #    set in prefs.
                #XXX This check for "Python" files should probably include
                #    python scripts, which might likely not have the
                #    extension: need to grow filetype-from-content smarts.
                files.append(path)

    def genScannableFiles(self, path=None, skipRareImports=False,
                          importableOnly=False):
        if path is None:
            path = self._getPath()
        searchedDirs = {}
        for dirname in path:
            files = []
            os.path.walk(dirname, self._findScannableFiles,
                         (files, searchedDirs, skipRareImports,
                          importableOnly))
            for file in files:
                yield file



class PerlImportHandler(ImportHandler):
    PATH_ENV_VAR = "PERL5LIB"

    def _shellOutForPath(self, compiler):
        import process
        sep = "--WomBa-woMbA--"
        argv = [compiler, "-e", "print join('%s', @INC);" % sep]
        env = dict(os.environ)
        if "PERL5LIB" in env: del env["PERL5LIB"]
        if "PERLLIB" in env: del env["PERLLIB"]

        p = process.ProcessOpen(argv, env=env)
        retval = p.wait()
        output = p.stdout.read()
        error = p.stderr.read()
        retval = p.wait()
        p.close()
        if retval:
            raise CodeIntelError("could not determine Perl import path: %s"
                                 % error)
        path = [os.path.normpath(d) for d in output.split(sep)]
        # cwd handled separately
        path = [p for p in path if p not in (os.curdir, os.getcwd())]
        return path

    def setCorePath(self, compiler=None, extra=None):
        if compiler is None:
            import which
            assert which._version_ >= (1, 0, 0)
            compiler = which.which("perl")
        self.corePath = self._shellOutForPath(compiler)

    #XXX These are only required for AutoComplete/CallTips and for
    #    "schedule scans of imports" functionality.
    #def findModule(self, cu, module, submodule, cwd=None):
    #    pass
    #def findModuleOnDisk(self, module, submodule, cwd=None):
    #    pass

    def _findScannableFiles(self,
                            (files, searchedDirs, skipTheseDirs, skipRareImports),
                            dirname, names):
        if sys.platform.startswith("win"):
            cpath = dirname.lower()
        else:
            cpath = dirname
        if cpath in searchedDirs:
            while names:
                del names[0]
            return
        else:
            searchedDirs[cpath] = 1
        for i in range(len(names)-1, -1, -1): # backward so can del from list
            path = os.path.join(dirname, names[i])
            if os.path.isdir(path):
                if os.path.normcase(path) in skipTheseDirs:
                    del names[i]
                elif skipRareImports and not ('A' <= names[i][0] <= 'Z'):
                    # Perl good practice dictates that all module directories
                    # begin with a capital letter. Therefore, we skip dirs
                    # that start with a lower case.
                    del names[i]
            elif os.path.splitext(names[i])[1] in (".pl", ".pm"):
                #XXX The list of extensions should be settable on
                #    the ImportHandler and Komodo should set whatever is
                #    set in prefs.
                #XXX This check for files should probably include
                #    scripts, which might likely not have the
                #    extension: need to grow filetype-from-content smarts.
                files.append(path)

    def genScannableFiles(self, path=None, skipRareImports=False,
                          importableOnly=False):
        if path is None:
            path = self._getPath()
        searchedDirs = {}
        for dirname in path:
            if dirname == os.curdir:
                # Do NOT traverse the common '.' element of @INC. It is
                # environment-dependent so not useful for the typical call
                # of this method.
                continue
            skipTheseDirs = [os.path.join(dirname, "auto")]
            skipTheseDirs = [os.path.normcase(d) for d in skipTheseDirs]
            files = []
            os.path.walk(dirname, self._findScannableFiles,
                         (files, searchedDirs, skipTheseDirs,
                          skipRareImports))
            for file in files:
                yield file


class PHPImportHandler(ImportHandler):
    def setCorePath(self, compiler=None, extra=None):
        #XXX To do this independent of Komodo this would need to do all
        #    the garbage that koIPHPInfoEx is doing to determine this. It
        #    might also require adding a "rcfile" argument to this method
        #    so the proper php.ini file is used in the "_shellOutForPath".
        #    This is not crucial now because koCodeIntel._Manager() handles
        #    this for us.
        if not self.corePath:
            raise CodeIntelError("Do not know how to determine the core "
                                 "PHP include path. 'corePath' must be set "
                                 "manually.")

    #XXX These are only required for AutoComplete/CallTips and for
    #    "schedule scans of imports" functionality.
    #def findModule(self, cu, module, submodule, cwd=None):
    #    pass
    #def findModuleOnDisk(self, module, submodule, cwd=None):
    #    pass

    def _findScannableFiles(self, (files, searchedDirs), dirname, names):
        if sys.platform.startswith("win"):
            cpath = dirname.lower()
        else:
            cpath = dirname
        if cpath in searchedDirs:
            while names:
                del names[0]
            return
        else:
            searchedDirs[cpath] = 1
        for i in range(len(names)-1, -1, -1): # backward so can del from list
            path = os.path.join(dirname, names[i])
            if os.path.isdir(path):
                pass
            elif os.path.splitext(names[i])[1] in (".php", ".inc"):
                #XXX The list of extensions should be settable on
                #    the ImportHandler and Komodo should set whatever is
                #    set in prefs.
                #XXX This check for files should probably include
                #    scripts, which might likely not have the
                #    extension: need to grow filetype-from-content smarts.
                files.append(path)

    def genScannableFiles(self, path=None, skipRareImports=False,
                          importableOnly=False):
        if path is None:
            path = self._getPath()
        searchedDirs = {}
        for dirname in path:
            if dirname == os.curdir:
                # Do NOT traverse '.' if it is in the include_path. Not sure
                # if this is at all common for PHP.
                continue
            files = []
            os.path.walk(dirname, self._findScannableFiles,
                         (files, searchedDirs))
            for file in files:
                yield file


class JavaScriptImportHandler(ImportHandler):
    def setCorePath(self, compiler=None, extra=None):
        self.corePath = []

    #XXX These are only required for AutoComplete/CallTips and for
    #    "schedule scans of imports" functionality.
    #def findModule(self, cu, module, submodule, cwd=None):
    #    pass
    #def findModuleOnDisk(self, module, submodule, cwd=None):
    #    pass

    def _findScannableFiles(self, (files, searchedDirs), dirname, names):
        if sys.platform.startswith("win"):
            cpath = dirname.lower()
        else:
            cpath = dirname
        if cpath in searchedDirs:
            while names:
                del names[0]
            return
        else:
            searchedDirs[cpath] = 1
        for i in range(len(names)-1, -1, -1): # backward so can del from list
            path = os.path.join(dirname, names[i])
            if os.path.isdir(path):
                pass
            elif os.path.splitext(names[i])[1] in (".js",):
                #XXX The list of extensions should be settable on
                #    the ImportHandler and Komodo should set whatever is
                #    set in prefs.
                #XXX This check for files should probably include
                #    scripts, which might likely not have the
                #    extension: need to grow filetype-from-content smarts.
                files.append(path)

    def genScannableFiles(self, path=None, skipRareImports=False,
                          importableOnly=False):
        if path is None:
            path = self._getPath()
        searchedDirs = {}
        for dirname in path:
            if dirname == os.curdir:
                # Do NOT traverse the common '.' element of @INC. It is
                # environment-dependent so not useful for the typical call
                # of this method.
                continue
            files = []
            os.path.walk(dirname, self._findScannableFiles,
                         (files, searchedDirs))
            for file in files:
                yield file


class TclImportHandler(ImportHandler):
    # Tcl _does_ have a TCLLIBPATH environment variable for specifying
    # import path elements, but parsing isn't straighforward -- it uses
    # Tcl-syntax -- so we don't bother to separate "envPath" and "corePath"
    # for Tcl.
    PATH_ENV_VAR = None

    def _shellOutForPath(self, compiler):
        import process
        argv = [compiler]
        p = process.ProcessOpen(argv)
        p.stdin.write("puts [join $auto_path \\n]")
        p.stdin.close()
        retval = p.wait()
        output = p.stdout.read()
        p.close()
        path = [os.path.normpath(line) for line in output.splitlines(0)]
        if path and (path[0] == "" or path[0] == os.getcwd()):
            del path[0] # cwd handled separately
        return path

    def setCorePath(self, compiler=None, extra=None):
        if compiler is None:
            import which
            assert which._version_ >= (1, 0, 0)
            compiler = which.which("tclsh")
        self.corePath = self._shellOutForPath(compiler)

    #XXX These are only required for AutoComplete/CallTips and for
    #    "schedule scans of imports" functionality.
    #def findModule(self, cu, module, submodule, cwd=None):
    #    pass
    #def findModuleOnDisk(self, module, submodule, cwd=None):
    #    pass

    def _findScannableFiles(self, (files, searchedDirs, skipRareImports),
                            dirname, names):
        if sys.platform.startswith("win"):
            cpath = dirname.lower()
        else:
            cpath = dirname
        if cpath in searchedDirs:
            while names:
                del names[0]
            return
        else:
            searchedDirs[cpath] = 1
        for i in range(len(names)-1, -1, -1): # backward so can del from list
            path = os.path.join(dirname, names[i])
            if os.path.isdir(path):
                pass
            elif os.path.splitext(names[i])[1] in (".tcl",):
                #XXX The list of extensions should be settable on
                #    the ImportHandler and Komodo should set whatever is
                #    set in prefs.
                #XXX This check for files should probably include
                #    scripts, which might likely not have the
                #    extension: need to grow filetype-from-content smarts.
                if skipRareImports and names[i] == "pkgIndex.tcl":
                    continue
                files.append(path)

    def genScannableFiles(self, path=None, skipRareImports=False,
                          importableOnly=False):
        if path is None:
            path = self._getPath()
        searchedDirs = {}
        for dirname in path:
            if dirname == os.curdir:
                continue
            files = []
            os.path.walk(dirname, self._findScannableFiles,
                         (files, searchedDirs, skipRareImports))
            for file in files:
                yield file



#---- CITDL processing stuff

class Object:
    """Represents a module or symbol object in the CIDB."""
    table = None
    id = None
    row = None

    def __init__(self, table, id, row=None, cu=None):
        # The 'row' can be passed in or a CIDB cursor, 'cu', can be passed
        # in and it will be retrieved.
        self.table = table
        self.id = id
        if row is None:
            cu.execute("SELECT * FROM %s WHERE id=%d", self.table, self.id)
            self.row = cu.fetchone()
        else:
            self.row = row

    def guid(self):
        return (self.table, self.id)

    def __repr__(self):
        details = ""
        if self.row:
            if self.table == "module":
                details = " %s" % self.row[M_NAME]
            elif self.table == "symbol":
                details = " %s %s" % (symbolType2Name(self.row[S_TYPE]),
                                      self.row[S_NAME])
            elif self.table == "import":
                details = ' '+str(self)
        klass = str(self.__class__).split('.')[-1]
        s = "<%s %s %s:%s>" % (klass, self.table, self.id, details)
        return s

    def __nonzero__(self):
        return self.table is not None and self.id is not None

    def getParentScope(self, cu, language):
        """Return a parent scope object for this scope.
        
            "cu" is an open CIDB cursor.
            "language" is the current scope file language.
        
        "Scoping" objects are modules, classes (namespaces are just a special
        kind of class) and functions. The CodeIntel system doesn't support
        lexical scopes created for if-, while-, for-, and plain brace-blocks.
        
        The parent of a class method is the class scope. The parent of a
        top-level class or function is its module. The parent of a module
        is the special built-in module (named "*"), if there is one.
        
        Returns None if there is no parent scope.
        """
        if self.table == "import":
            raise CodeIntelError("shouldn't be calling getParentScope on an"
                                 "Import instance")
        elif self.table == "module" and self.row[M_NAME] == "*":
            scope = None  # no parent of the built-in module
        elif self.table == "module":
            # Look for a built-in module for this language.
            cu.execute("SELECT module.* FROM module, file, language "
                       "WHERE module.name='*' AND language.name=%s "
                       "AND language.id=file.language_id "
                       "AND module.file_id=file.id", language)
            if cu.rowcount:
                row = cu.fetchone()
                scope = Scope("module", row[M_ID], row)
            else:
                scope = None
        elif not self.row[S_SCOPE_ID]:
            module_id = int(self.row[S_MODULE_ID])
            cu.execute("SELECT * FROM module WHERE id=%d", int(module_id))
            if not cu.rowcount:
                raise CIDBError("id %r not found in module table" % module_id)
            scope = Scope("module", module_id, cu.fetchone())
        else:
            symbol_id = int(self.row[S_SCOPE_ID])
            cu.execute("SELECT * FROM symbol WHERE id=%d", int(symbol_id))
            if not cu.rowcount:
                raise CIDBError("id %r not found in symbol table" % symbol_id)
            scope = Scope("symbol", symbol_id, cu.fetchone())
        return scope

    def getType(self):
        if self.table == "module":
            return "module"
        else:
            return self.row[S_TYPE]

class Scope(Object):
    pass

class Import(Object):
    """Technically I said above that an 'Object' instance is limited to
    module or symbol entries. In general that is true, but for a short while
    in processing 'Import' objects can exist. They are alway resolved away
    to module or symbol entries, though.
    """
    def __str__(self):
        row = self.row
        module, symbol, alias = row[I_MODULE], row[I_SYMBOL], row[I_ALIAS]
        if alias and symbol:
            s = "from %(module)s import %(symbol)s as %(alias)s" % locals()
        elif alias:
            s = "import %(module)s as %(alias)s" % locals()
        elif symbol:
            s = "from %(module)s import %(symbol)s" % locals()
        else:
            s = "import %(module)s" % locals()
        return s

    def getType(self):
        return "import"


def getScopeForFileAndLine(cu, file, line):
    """Return a Scope instance for the given position.
    
        "cu" is an open CIDB cursor.
        "file" is path to the file (not-canonicalized).
        "line" is the (1-based) line number at which to determine the
            scope.
    """
    file_id = _getFileId(cu, file)

    # Get all "scoping" symbols, ordered by line, for this file_id and
    # find the appropriate bounding symbol (note that "lineend" is
    # currently an optional field.
    cu.execute("SELECT * FROM symbol WHERE file_id=%d "
               "AND type IN (%d, %d, %d) ORDER BY line",
               file_id, ST_CLASS, ST_FUNCTION, ST_INTERFACE)
    symbolRow = None
    for row in cu.fetchall():
        if (row[S_LINE] <= line and
            (row[S_LINEEND] is None or line <= row[S_LINEEND])):
            symbolRow = row
        elif row[S_LINE] > line:
            break # we've already gone past the given "line"
    if symbolRow:
        scope = Scope("symbol", symbolRow[S_ID], symbolRow)
        return scope
    
    # We are at module-level, return id of the appropriate module.
    cu.execute("SELECT * FROM module WHERE file_id=%d ORDER BY line",
               file_id)
    if not cu.rowcount:
        raise CodeIntelError("there are no CIDB module entries for "
                             "file %d: '%s'" % (file_id, file))
    moduleRow = None
    for row in cu.fetchall():
        if row[M_LINE] <= line:
            moduleRow = row
        elif row[M_LINE] > line:
            break # we've gone past the given "line"
    if not moduleRow:
        #XXX Why is this being hit so frequently? Figure it out.
        raise CodeIntelError("unexpectedly, no module rows for file "
                             "%d correspond to line %d"
                             % (file_id, line))
    scope = Scope("module", moduleRow[M_ID], moduleRow)
    return scope


def getModuleScopeFromFileAndLine(cu, file, line):
    """Return a Scope instance representing the current module."""
    file_id = _getFileId(cu, file)

    cu.execute("SELECT * FROM module WHERE file_id=%d ORDER BY line", file_id)
    if not cu.rowcount:
        raise CodeIntelError("there are no CIDB module entries for file %d"
                             % file_id)
    moduleRow = None
    for row in cu.fetchall():
        if not moduleRow:
            # Always fallback to first module in file: common case.
            moduleRow = row
        if row[M_LINE] <= line:
            moduleRow = row
        elif row[M_LINE] > line:
            break # we've gone past the given "line"

    scope = Scope("module", moduleRow[M_ID], moduleRow)
    return scope


class Context:
    # May want to cache data for the current file rather than starting
    # from zero for each getMembers/getCallTips in that file.
    # c.f. Memoize: http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/52201
    def __init__(self, manager, language, path, line):
        """Create a CITDL evaluation Context.
        
            "manager" is the controlling codeintel.Manager instance.
            "language" is the file language in which to evaluate.
            "path" is the path to the file.
            "line" is the (1-based) trigger line. The path and line together
                typically determine the scope of evaluation.
        """
        log.debug("Context(manager, language=%r, path=%r, line=%r)",
                  language, path, line)
        self.manager = manager
        self.language = language
        self.path = path
        self.line = line
        self.file_id = None # determined lazily
        self.module_id = None # determined lazily
        self.notes = []
        # An internal list of object on whom members have been retrieved.
        # This is used to guard against infinite loops (see
        # test/test_eval_citdl/infinity.py). It is reset for each
        # getMembers() or getCallTips().
        self.__hitObjects = None

    def getMembers(self, cu, expr, scopeTable=None, scopeId=0):
        """Get the members for the given CITDL expression.
        
            "cu" is an open CIDB cursor
            "expr" is the CITDL expression
            "scopeTable" and "scopeId" (optional) can be passed it to
                express the scope more accurately than "path" and "line" --
                if, say, the caller is adjusting for recent edits not in the
                database.
        """
        self.log("getMembers for '%s' at %s:%d", expr,
                 os.path.basename(self.path), self.line)
        if scopeTable:
            scope = Scope(scopeTable, scopeId, cu=cu)
        else:
            scope = getScopeForFileAndLine(cu, self.path, self.line)
        self.__hitObjects = {}
        members = self._getMembers(cu, expr, scope)
        self.log("    members: %r", members)
        return members

    def getCallTips(self, cu, expr, scopeTable=None, scopeId=0):
        """Get the members for the given CITDL expression.
        
            "cu" is an open CIDB cursor
            "expr" is the CITDL expression
            "scopeTable" and "scopeId" (optional) can be passed it to
                express the scope more accurately than "path" and "line" --
                if, say, the caller is adjusting for recent edits not in the
                database.
        """
        self.log("getCallTips for '%s' at %s:%d", expr,
                 os.path.basename(self.path), self.line)
        if scopeTable:
            scope = Scope(scopeTable, scopeId, cu=cu)
        else:
            scope = getScopeForFileAndLine(cu, self.path, self.line)
        self.__hitObjects = {}
        calltips = self._getCallTips(cu, expr, scope)
        self.log("    calltips: %r", calltips)
        return calltips

    def log(self, msg, *args):
        self.notes.append(msg % args)
        log.info(msg, *args)
    def warn(self, msg, *args):
        self.notes.append(msg % args)
        log.info(msg, *args) # don't want to alarm user's with warn-level log output
    def error(self, msg, *args):
        self.notes.append(msg % args)
        log.error(msg, *args)

    __moduleScope = None
    def _getModuleScope(self, cu):
        """Make the module scope a singleton as an optimization."""
        if self.__moduleScope is None:
            self.__moduleScope = getModuleScopeFromFileAndLine(cu, self.path,
                                                               self.line)
        return self.__moduleScope

    def _parseScope(self, cu, expr, scope=None):
        """Determine and return the appropriate scope.

        The scope is determined like this:
        - the "expr" expression can specify it, e.g. in
          "sys.version_info Foo.bar" the scope is "Foo.bar"
        - otherwise, if "scope" is specified, then that is it
        - otherwise, module-level scope is implied.

        This returns (1) the CITDL <symbol-ref> part, with any possible
        <start-scope> part removed; and, (2) the scope -- a Scope instance.
        """
        if ' ' in expr:
            try:
                symbolRef, scopeStr = expr.split()
            except ValueError, ex:
                raise CITDLError("invalid CITDL expression: can only have "
                                 "two parts: %r" % expr)
            startScope = self._resolveStartScope(cu, scopeStr)
        elif scope is not None:
            symbolRef = expr
            startScope = scope
        else:
            # Get the module-level scope.
            symbolRef = expr
            startScope = self._getModuleScope(cu)
        retval = (symbolRef, startScope)
        log.debug("Context._parseScope(expr=%r, scope=%r): return %r",
                  expr, scope, retval)
        return retval

    def _resolveStartScope(self, cu, scopeStr):
        """Resolve the given <start-scope> string part of a CITDL expr.

        This is a simpler expression than the <symbol-ref> part of a full
        CITDL expression: '()' tokens are illegal and only scoping symbols
        (classes, interfaces, functions) and modules are valid. I.e. no
        imports and no variables/arguments.
        
        Returns a Scope instance for the resolved scope or raises a
        CodeIntelError.
        """
        if '(' in scopeStr or ')' in scopeStr:
            raise CITDLError("illegal CITDL syntax: '()' not allowed in "
                             "<start-scope> part: '%s'" % scopeStr)

        # <start-scope> parts are always based at module-scope
        scope = self._getModuleScope(cu)
        for token in scopeStr.split('.'):
            if scope.table == "module":
                cu.execute("SELECT * FROM symbol "
                           " WHERE module_id=%d AND NOT scope_id "
                           "   AND name=%s AND type IN (%d, %d, %d)",
                           scope.id, token, ST_CLASS, ST_FUNCTION, ST_INTERFACE)
            else: # scope.table == "symbol"
                cu.execute("SELECT * FROM symbol "
                           " WHERE scope_id=%d AND name=%s "
                           "   AND type IN (%d, %d, %d)",
                           scope.id, token, ST_CLASS, ST_FUNCTION, ST_INTERFACE)
            if not cu.rowcount:
                raise CodeIntelError("could not resolve '%s': could not "
                                     "find '%s' on %r"
                                     % (scopeStr, token, scope))
            elif cu.rowcount > 1:
                raise CodeIntelError("could not resolve '%s': '%s' is "
                                     "ambiguous on %r"
                                     % (scopeStr, token, scope))
            row = cu.fetchone()
            scope = Scope("symbol", row[S_ID], row)

        self.log("    start scope '%s' resolves to %r", scopeStr, scope)
        return scope

    def _eval(self, cu, expr, scope=None):
        """XXX 'splain
        """
        # A CITDL expression is of the following form:
        #   <symbol-ref>[ in <start-scope>]
        # where,
        #   <symbol-ref> is a chain of identifiers, '.' for attribute access,
        #       and '()' for calling, e.g. "os.environ.get().split()"
        #   <start-scope> is a chain "scoping" identifiers and '.' for
        #       attribute access. "Scoping" here means that only symbols that
        #       create new scopes are relevant, i.e. classes, functions and
        #       interfaces but not arguments and variables. A start scope
        #       is always in the current module so a module reference is
        #       not relevant either.
        #XXX The "in" keyword is newly proposed and is not currently used.
        symbolRef, startScope = self._parseScope(cu, expr, scope)
        
        # Try to resolve the CITDL expression in the current scope,
        # then try to the parent scope, ... then try the module-level scope,
        # (XXX: then try to "builtins-scope" (a.k.a. the "*" module)).
        scope = startScope
        while 1:
            hits = self._evalAtScope(cu, symbolRef, scope)
            if hits:
                # Guard against an infinite loop.
                newhits = []
                for hit in hits:
                    guid = hit.guid()
                    if guid in self.__hitObjects:
                        log.info("    possible infinite loop on %r: "
                                 "dropping it", hit)
                    else:
                        self.__hitObjects[hit.guid()] = 1
                        newhits.append(hit)
                return newhits
            else:
                scope = scope.getParentScope(cu, self.language)
            if not scope:
                raise CodeIntelError("could not resolve '%s' in %r"
                                     % (symbolRef, startScope))

    def _evalAtScope(self, cu, expr, scope):
        """Attempt to resolve the given CITDL symbol reference at the
        given scope.

            "cu" is an open CIDB cursor.
            "expr" is the <symbol-ref> part of a CITDL expression.
            "scope" is the Scope on which to search.

        Returns a list of Object instances (one for each hit) or an empty list
        if there were no hits at this scope. It can raise a CodeIntelError for
        limitations or other errors.
        """
        hits = [scope]
        tokenizer = re.compile("(\.|\(\))")
        isFirstAttrToken = 1
        for token in tokenizer.split(expr):
            if token == '.':
                continue
            elif token == '()':
                raise CodeIntelError("cannot yet evaluate CITDL expressions "
                                     "using '()': '%s'" % expr)
            else:
                attrHits = []
                for hit in hits:
                    attrHits += self._resolveGetattr(cu, hit, token,
                                                     isFirstAttrToken)
                    isFirstAttrToken = 0
                hits = attrHits
        return hits

    def _resolveImport(self, cu, impobj, name, first=0):
        """Resolve the given import to a symbol or module object.

            "obj" is an Import instance, i.e. an import table instance
            "name" is the name of the symbol to find on "obj".
            "first" is a boolean indicating if this is the first symbol name
                in the CITDL expression (it matters for functions).

        Returns a list of possible hits for the import lookup. If the module
        or specifically named symbol (i.e. not '*') cannot be found in the
        database, an error is raised. If a symbol cannot be found in a
        '*'-type import, that is fine, the empty list is returned.
        """
        hits = []

        module = impobj.row[I_MODULE]
        symbol = impobj.row[I_SYMBOL]
        alias = impobj.row[I_ALIAS]

        handler = self.manager.getImportHandler(self.language)
        cwd = os.path.dirname(self.path)
        mrow, kind = handler.findModule(cu, module, symbol, cwd=cwd)
        if (kind == "module" and not symbol) or kind == "submodule":
            # Covers:   import mod
            #           import mod as alias
            #           from pkg import mod
            #           from pkg import mod as alias
            hits = [ Object("module", mrow[0], mrow) ]
        else:
            # Covers:   from mod import symbol
            #           from mod import symbol as alias
            #           from mod import *
            # Have to lookup the appropriate symbol on the found module.
            if alias:
                symbolName = symbol
            else:
                symbolName = name
            cu.execute("SELECT * FROM symbol "
                       "WHERE module_id=%d AND NOT scope_id AND name=%s",
                       mrow[0], symbolName)
            for row in cu.fetchall():
                hits.append( Object("symbol", row[S_ID], row) )
            cu.execute("SELECT * FROM import "
                       "WHERE module_id=%d AND NOT scope_id "
                       "AND (alias=%s OR "
                       "     (NOT alias AND "
                       "      (symbol=%s OR "
                       "       (NOT symbol AND module=%s))))",
                       mrow[0], symbolName, symbolName, symbolName)
            for irow in cu.fetchall():
                iobj = Import("import", irow[I_ID], irow)
                ihits = self._resolveImport(cu, iobj, symbolName, first)
                hits += ihits
        self.log("    does '%s' import '%s'? %s", impobj, name,
                 hits and "yes" or "no")
        if log.isEnabledFor(logging.INFO) and hits:
            for hit in hits:
                self.log("        %r", hit)
        return hits

    def _resolveGetattr(self, cu, obj, name, first=0):
        """Resolve the given attribute/symbol lookup.

            "obj" is an Object instance, i.e. a module or symbol row reference
            "name" is the name of the symbol to find on "obj".
            "first" is a boolean indicating if this is the first symbol name
                in the CITDL expression (it matters for functions).
        
        Returns a list of possible hits for the attr lookup or the empty
        list if it could not be resolved. It can raise a CodeIntelError
        for limitations or other errors.
        """
        hits = {} # use dict to eliminate dupes, maps Object.guid() to Object
        if obj.table == "module":
            symbolType = None
        else:
            symbolType = int(obj.row[S_TYPE])

        if symbolType in (ST_VARIABLE, ST_ARGUMENT):
            cu.execute("SELECT citdl FROM type WHERE symbol_id=%d", obj.id)
            typeInferences = [row[0] for row in cu.fetchall()]
            self.log("    type inferences for %s symbol %d: %r",
                     symbolType2Name(symbolType), obj.id, typeInferences)
            for expr in typeInferences:
                for o in self._eval(cu, expr):
                    for h in self._resolveGetattr(cu, o, name, first):
                        hits[h.guid()] = h
        elif not first and symbolType == ST_FUNCTION:
            # Function arguments and internal variables are NOT attributes of
            # the function. Therefore, these should only resolve if they are
            # the first part of a CITDL expression. E.g., given this function:
            #   def foo(arg1, arg2):
            #       var1 = 42
            #       #...
            # "arg1" and "var1" _in_ the function should resolve, but
            # "foo.var1" and "foo.arg1" should NOT.
            pass
        else:
            # Check imports for a match.
            if obj.table == "module":
                cu.execute("SELECT * FROM import "
                           "WHERE module_id=%d AND NOT scope_id", obj.id)
            else:
                cu.execute("SELECT * FROM import WHERE scope_id=%d", obj.id)
            for irow in cu.fetchall():
                iname = irow[I_ALIAS] or irow[I_SYMBOL] or irow[I_MODULE]
                if iname == name or iname == "*":
                    impobj = Import("import", irow[I_ID], irow)
                    try:
                        for o in self._resolveImport(cu, impobj, name):
                            hits[o.guid()] = o
                    except CodeIntelError, ex:
                        if iname == "*":
                            # Failure to resolve a "*" import shouldn't be
                            # fatal. There is a decent chance the symbol
                            # comes from elsewhere. Otherwise an errant
                            # "*"-import would kill all intellisense.
                            self.warn("    error resolving %r: ignoring",
                                      impobj)
                        else:
                            raise

            # Check locally defined symbols for a match.
            if obj.table == "module":
                cu.execute("SELECT * FROM symbol "
                           "WHERE module_id=%d AND NOT scope_id AND name=%s",
                           obj.id, name)
            else:
                cu.execute("SELECT * FROM symbol WHERE scope_id=%d AND name=%s",
                           obj.id, name)
            for row in cu.fetchall():
                o = Object("symbol", row[S_ID], row)
                hits[o.guid()] = o
            
            # Check base classes and interfaces (if relevant).
            if symbolType == ST_CLASS and not cu.rowcount:
                cu.execute("SELECT id, name FROM base WHERE symbol_id=%d",
                           obj.id)
                bases = cu.fetchall()
                if bases:
                    self.log("    bases for %s symbol %d: %r", symbolType,
                             obj.id, [b[1] for b in bases])
                    objScope = obj.getParentScope(cu, self.language)
                    for baseId, baseName in bases:
                        cu.execute("SELECT citdl FROM type WHERE base_id=%d",
                                   baseId)
                        baseTypeInferences = [row[0] for row in cu.fetchall()]
                        for expr in baseTypeInferences:
                            try:
                                for o in self._eval(cu, expr, objScope):
                                    try:
                                        for h in self._resolveGetattr(cu, o, name, first):
                                            hits[h.guid()] = h
                                    except CodeIntelError, ex:
                                        self.warn("    error getting '%s' attr "
                                                  "on base %s %r: %s: ignoring",
                                                  name, o.getType(), o, ex)
                            except CodeIntelError, ex:
                                self.warn("    error evaluating base '%s': "
                                          "%s: ignoring", expr, ex)

        hits = hits.values()
        self.log("    is '%s' accessible on %r? %s", name, obj,
                 hits and "yes" or "no")
        if log.isEnabledFor(logging.INFO) and hits:
            for hit in hits:
                self.log("        %r", hit)
        return hits
    
    def _getSymbolMembers(self, cu, symbol):
        members = []
        symbolType = int(symbol.row[S_TYPE])
        if symbolType == ST_FUNCTION:
            # Function arguments and internal variables should NOT be shown
            # in an autocomplete list for the function. Those would be
            # function attributes, if we supported them -- which we currently
            # don't. (Do many languages support/use them -- Python does, but
            # their use is rare, I think).
            pass
        elif symbolType in (ST_CLASS, ST_INTERFACE):
            cu.execute("SELECT name, type FROM symbol "
                       "WHERE scope_id=%d", symbol.id)
            members = {}
            for name, type in cu.fetchall():
                #XXX Will want to pass 'type' eventually to
                #    put proper icons in autocomplete UI.
                members[name] = 1
            # Add base class members.
            cu.execute("SELECT id, name FROM base WHERE symbol_id=%d",
                       symbol.id)
            bases = [row for row in cu.fetchall()]
            if bases:
                self.log("    bases for %s symbol %d: %r", symbolType,
                         symbol.id, [b[1] for b in bases])
                symbolScope = symbol.getParentScope(cu, self.language)
                for baseId, baseName in bases:
                    cu.execute("SELECT citdl FROM type WHERE base_id=%d",
                               baseId)
                    baseTypeInferences = [row[0] for row in cu.fetchall()]
                    self.log("    type inferences for base '%s' (%d): %r",
                             baseName, baseId, baseTypeInferences)
                    baseMembers = []
                    for expr in baseTypeInferences:
                        try:
                            baseMembers += self._getMembers(cu, expr,
                                                            symbolScope)
                        except CodeIntelError, ex:
                            self.warn("    error getting members from "
                                      "base '%s': %s: ignoring", expr, ex)
                    self.log("    base '%s' members: %r", baseName, baseMembers)
                    for baseMember in baseMembers:
                        members[baseMember] = 1
            members = members.keys()
        elif symbolType in (ST_VARIABLE, ST_ARGUMENT):
            cu.execute("SELECT citdl FROM type WHERE symbol_id=%d",
                       int(symbol.id))
            typeInferences = [row[0] for row in cu.fetchall()]
            self.log("    type inferences for %s symbol %d: %r",
                     symbolType2Name(symbolType), symbol.id, typeInferences)
            for expr in typeInferences:
                members += self._getMembers(cu, expr)
        else:
            raise CodeIntelError("cannot get members from CITDL that "
                                 "resolves to a '%(type)s' symbol: "
                                 "%(name)s (id=%(id)d)" % symbol.row)
        return members

    def _getModuleMembers(self, cu, module):
        members = []
        cu.execute("SELECT name FROM symbol WHERE module_id=%d AND NOT scope_id",
                   module.id)
        for (name,) in cu.fetchall():
            members.append(name)

        cu.execute("SELECT * FROM import "
                   "WHERE module_id=%d AND NOT scope_id",
                   module.id)
        for irow in cu.fetchall():
            iobj = Import("import", irow[I_ID], irow)
            members += self._getImportedMembers(cu, iobj)

        return members

    def _getImportedMembers(self, cu, impobj):
        row = impobj.row
        module, symbol, alias = row[I_MODULE], row[I_SYMBOL], row[I_ALIAS]
        if alias:
            members = [alias]
        elif symbol == "*":
            log.info("    warning: don't yet know how to get '*' import: "
                     "'from %s import *'", module)
            members = []
        elif symbol:
            members = [symbol]
        elif module:
            members = [module]
        return members

    def _getMembers(self, cu, expr, scope=None):
        """Get the members for the given CITDL expression and scope.
        
            "cu" is an open CIDB cursor.
            "expr" is a CITDL expression.
            "scope" (optional) is a Scope instance specifying the start
                scope. Not that a scope specified in "expr" might still
                override this (see comment for _eval()).
        """
        members = []
        hits = self._eval(cu, expr, scope)
        for hit in hits:
            if hit.table == "module":
                members += self._getModuleMembers(cu, hit)
            elif hit.table == "symbol":
                members += self._getSymbolMembers(cu, hit)
            else:
                raise CodeIntelError("cannot get members from CITDL that "
                                     "resolves to a '%s': %r"
                                     % (hit.table, expr))
        return members

    def _trimDocForCalltip(self, doc):
        """Reduce the given doc string for inclusion in a calltip.
        
            "doc" is the full doc-string to reduce.
        
        The trick is to come up with something that is more useful that just
        the first non-blank line, but not a really bad for edge cases.
        Chosen algorithm:
        - trim leading blank lines
        - find the indentation of the first non-blank line
        - accumulate lines until the indentation changes or there is a blank
          line (or perhaps should stop just on a blank line)
        - trim the leadning indentation from all lines
        - limit to 5 lines
        """
        ## Style: first non-blank line
        #for line in doc.splitlines(0):
        #    if line.strip():
        #        trimmed = line.strip()
        #        break

        # Style: leading non-blank lines with same indent
        MAX_LINES = 5
        trimmed = ""
        trimmedLines = 0
        firstIndent = ""
        for line in doc.splitlines(0):
            if trimmedLines >= MAX_LINES:
                break
            if not trimmed and not line.strip(): # leading blank line
                continue
            elif not trimmed: # first non-blank line
                for ch in line:
                    if ch in " \t":
                        firstIndent += ch
                    else:
                        break
                trimmed += line.strip()
                trimmedLines += 1
            elif not line.strip(): # trailing blank line
                break
            else:
                indent = ""
                for ch in line:
                    if ch in " \t":
                        indent += ch
                    else:
                        break
                if indent != firstIndent: # different indent on this line
                    break
                trimmed += "\n"+line.rstrip()[len(indent):]
                trimmedLines += 1
        
        return trimmed

    def _getSymbolCallTips(self, cu, symbol):
        symbolType = int(symbol.row[S_TYPE])
        calltips = []

        if symbolType == ST_FUNCTION:
            name = symbol.row[S_NAME]
            cu.execute("SELECT * FROM symbol WHERE scope_id=%d "
                       "AND type=%d ORDER BY id", symbol.id, ST_ARGUMENT)
            args = [row[S_SIGNATURE] or row[S_NAME] for row in cu.fetchall()]
            #XXX Hack to remove 'self' from method calltips for Python.
            #    Technically 'self' should remain if an unbound method or
            #    static method is being called, but we haven't
            if self.language == "Python" and args and args[0] == "self":
                cu.execute("SELECT type FROM symbol WHERE id=%d",
                           symbol.row[S_SCOPE_ID])
                if cu.rowcount and int(cu.fetchone()[0]) == ST_CLASS:
                    del args[0]
            calltip = "%s(%s)" % (symbol.row[S_NAME], ', '.join(args))
            doc = symbol.row[S_DOC]
            if doc:
                calltip += "\n" + self._trimDocForCalltip(doc)
            calltips.append(calltip)

        elif symbolType == ST_CLASS:
            # The calltip is the signature of the class constructor(s).
            # In the ideal world we would:
            # - find all constructor methods (including those in base
            #   classes);
            # - eliminate those that are overridden (based on
            #   language-specific details on when this occurs, e.g.: method
            #   name match in Python, signature match in C++); and
            # - return them all
            # However, because we currently only show on calltip in the UI
            # we can either only return one or not bother culling overridden
            # constructor -- which ever is easier.
            #XXX Presuming that we are not getting false positive in the
            #    attribute comparison (seeing as '_' is a SQL LIKE wildcard
            #    character).
            classDoc = symbol.row[S_DOC]
            cu.execute("SELECT * FROM symbol WHERE scope_id=%d "
                       "   AND type=%d AND attributes LIKE '%%__ctor__%%'",
                       symbol.id, ST_FUNCTION)
            ctorSymbols = cu.fetchall()
            foundSingleLocalCtor = len(ctorSymbols)==1
            if not ctorSymbols:
                # Optimization: While we only show one possible ctor in the
                # UI and we found one already, we only search the base
                # classes if we couldn't find a ctor on the class itself.
                cu.execute("SELECT id, name FROM base WHERE symbol_id=%d",
                           symbol.id)
                bases = cu.fetchall()
                if bases:
                    self.log("    bases for %s symbol %d: %r", symbolType,
                             symbol.id, [b[1] for b in bases])
                    for baseId, baseName in bases:
                        cu.execute("SELECT citdl FROM type WHERE base_id=%d",
                                   baseId)
                        baseTypeInferences = [row[0] for row in cu.fetchall()]
                        for expr in baseTypeInferences:
                            try:
                                for hit in self._eval(cu, expr):
                                    cu.execute("SELECT * FROM symbol "
                                               " WHERE scope_id=%d "
                                               "   AND type=%d "
                                               "   AND attributes LIKE '%%__ctor__%%'",
                                               hit.id, ST_FUNCTION)
                                    ctorSymbols += cu.fetchall()
                                    if ctorSymbols: break  # Optimization
                            except CodeIntelError, ex:
                                self.warn("    error evaluating base '%s': "
                                          "%s: ignoring", expr, ex)
                            if ctorSymbols: break  # Optimization
                        if ctorSymbols: break  # Optimization
            for ctor in ctorSymbols:
                id, name = ctor[S_ID], ctor[S_NAME]
                cu.execute("SELECT * FROM symbol WHERE scope_id=%d "
                           "AND type=%d ORDER BY id", id, ST_ARGUMENT)
                args = [row[S_SIGNATURE] or row[S_NAME] for row in cu.fetchall()]
                #XXX Hack to remove 'self' from method calltips for Python.
                #    Technically 'self' should remain if an unbound method or
                #    static method is being called, but we haven't
                if self.language == "Python" and args and args[0] == "self":
                    del args[0]
                #XXX Might want to use a scopesep other than '.' for other
                #    languages. Should put that in one place.
                calltip = "%s.%s(%s)" % (ctor[S_SCOPE], name, ', '.join(args))
                # Favour the _class_ documentation over the constructor
                # function documentation if a single ctor was found locally
                # (i.e. in this classdef rather than in base classes).
                doc = foundSingleLocalCtor and classDoc or ctor[S_DOC]
                if doc:
                    calltip += "\n" + self._trimDocForCalltip(doc)
                calltips.append(calltip)

        elif symbolType in (ST_ARGUMENT, ST_VARIABLE):
            cu.execute("SELECT citdl FROM type WHERE symbol_id=%d", symbol.id)
            typeInferences = [row[0] for row in cu.fetchall()]
            self.log("    type inferences for %s symbol %d: %r",
                     symbolType2Name(symbolType), symbol.id, typeInferences)
            for expr in typeInferences:
                calltips += self._getCallTips(cu, expr)

        else:
            raise CodeIntelError("cannot get calltips from CITDL that "
                                 "resolves to a '%(type)s' symbol: "
                                 "%(name)s (id=%(id)d)" % symbol.row)
        return calltips

    def _getCallTips(self, cu, expr, scope=None):
        """Get the calltips for the given CITDL expression and scope.
        
            "cu" is an open CIDB cursor.
            "expr" is a CITDL expression.
            "scope" (optional) is a Scope instance specifying the start
                scope. Not that a scope specified in "expr" might still
                override this (see comment for _eval()).
        """
        calltips = []
        hits = self._eval(cu, expr, scope)
        for hit in hits:
            if hit.table == "module":
                pass # modules are not callable
            elif hit.table == "symbol":
                calltips += self._getSymbolCallTips(cu, hit)
            else:
                raise CodeIntelError("cannot get calltips from CITDL that "
                                     "resolves to a '%s': %r"
                                     % (hit.table, expr))
        return calltips



#---- mainline self-test

if __name__ == '__main__':
    def _test():
        # To run from the command line: python build.py _test
        # This requires Python >=2.1 to run.
        import doctest, citdl
        return doctest.testmod(citdl)
    def _test_TclImportHandler():
        importer = TclImportHandler()
        importer.setCorePath()
        print importer._getPath()
        path = None
        #path = [r"D:\trentm\as\Apps\Komodo-devel\src"]
        for file in importer.genScannableFiles(path, skipRareImports=True):
            print "scannable Tcl file: '%s'" % file
    #_test()
    _test_TclImportHandler()

